# Lint as: python3
# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `privacy.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
import jax
import jax.numpy as jnp

from optax._src import privacy


class DifferentiallyPrivateAggregateTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.batch_size = 8
    self.params = {'key_a': (jnp.zeros((2, 3, 4)), jnp.zeros([])),
                   'key_b': jnp.zeros((6, 7))}
    # Example `i`'s grads are full of `i`s. Important to include 0 to ensure
    # there are no divisons by 0 (e.g. in norm clipping)
    a = jnp.arange(self.batch_size)
    self.per_eg_grads = jax.tree_map(
        lambda p: jnp.moveaxis(a * jnp.ones(p.shape+(self.batch_size,)), -1, 0),
        self.params)

  @chex.all_variants
  def test_no_privacy(self):
    """l2_norm_clip=MAX_FLOAT32 and noise_multiplier=0 should recover SGD."""
    dp_agg = privacy.differentially_private_aggregate(
        l2_norm_clip=jnp.finfo(jnp.float32).max,
        noise_multiplier=0.,
        seed=0)
    state = dp_agg.init(self.params)
    update_fn = self.variant(dp_agg.update)
    mean_grads = jax.tree_map(lambda g: g.mean(0), self.per_eg_grads)

    for _ in range(3):
      updates, state = update_fn(self.per_eg_grads, state)
      chex.assert_tree_all_close(updates, mean_grads)

  @chex.all_variants
  @parameterized.parameters(0.5, 10.0, 20.0, 40.0, 80.0)
  def test_clipping_norm(self, l2_norm_clip):
    dp_agg = privacy.differentially_private_aggregate(
        l2_norm_clip=l2_norm_clip,
        noise_multiplier=0.,
        seed=42)
    state = dp_agg.init(self.params)
    update_fn = self.variant(dp_agg.update)

    # Shape of the three arrays below is (self.batch_size, )
    norms = [jnp.linalg.norm(g.reshape(self.batch_size, -1), axis=1)
             for g in jax.tree_leaves(self.per_eg_grads)]
    global_norms = jnp.linalg.norm(jnp.stack(norms), axis=0)
    divisors = jnp.maximum(global_norms / l2_norm_clip, 1.)
    # Since the values of all the parameters are the same within each example,
    # we can easily compute what the values should be:
    expected_val = jnp.mean(jnp.arange(self.batch_size) / divisors)
    expected_tree = jax.tree_map(
        lambda p: jnp.broadcast_to(expected_val, p.shape), self.params)

    for _ in range(3):
      updates, state = update_fn(self.per_eg_grads, state, self.params)
      chex.assert_tree_all_close(updates, expected_tree)

  @chex.all_variants
  @parameterized.parameters((3.0, 2.0), (1.0, 5.0), (100.0, 4.0), (1.0, 90.0))
  def test_noise_multiplier(self, l2_norm_clip, noise_multiplier):
    """Standard dev. of noise should be l2_norm_clip * noise_multiplier."""
    dp_agg = privacy.differentially_private_aggregate(
        l2_norm_clip=l2_norm_clip,
        noise_multiplier=noise_multiplier,
        seed=1337)
    state = dp_agg.init(None)
    update_fn = self.variant(dp_agg.update)
    expected_std = l2_norm_clip * noise_multiplier

    grads = [jnp.ones((1, 100, 100))]  # batch size 1
    for _ in range(3):
      updates, state = update_fn(grads, state)
      chex.assert_tree_all_close(expected_std,
                                 jnp.std(updates[0]),
                                 atol=0.1 * expected_std)

  def test_aggregated_updates_as_input_fails(self):
    """Expect per-example gradients as input to this transform."""
    dp_agg = privacy.differentially_private_aggregate(l2_norm_clip=0.1,
                                                      noise_multiplier=1.1,
                                                      seed=2021)
    state = dp_agg.init(self.params)
    mean_grads = jax.tree_map(lambda g: g.mean(0), self.per_eg_grads)
    with self.assertRaises(ValueError):
      dp_agg.update(mean_grads, state, self.params)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A basic MNIST example using the Adam optimizer and lookahead wrapper."""
import functools
from typing import Any, Callable, Iterable, Mapping, Sequence, Tuple

from absl import app
from absl import flags
import haiku as hk
import jax
from jax import random
import jax.numpy as jnp
import optax
from optax.examples import datasets

flags.DEFINE_float('learning_rate', 0.002, 'Learning rate of fast optimizer.')
flags.DEFINE_float('slow_step_size', 0.5,
                   'Step size for slow lookahead parameters.')
flags.DEFINE_integer(
    'sync_period', 5,
    'Number of fast optimizer steps to take before synchronizing parameters.')
flags.DEFINE_multi_integer(
    'hidden_dims', [1000, 1000], 'Hidden dimensions of the dense layer. '
    'Each use of the flag adds another layer with the specified dimension.')
flags.DEFINE_integer('batch_size', 128, 'Batch size to use for training.')
flags.DEFINE_integer('n_epochs', 5, 'Number of epochs to train for.')
flags.DEFINE_integer('seed', 1, 'Random seed.')
FLAGS = flags.FLAGS


def categorical_crossentropy(logits: jnp.ndarray,
                             labels: jnp.ndarray) -> jnp.ndarray:
  losses = -jnp.sum(labels * jax.nn.log_softmax(logits), axis=1)
  return jnp.mean(losses)


def accuracy(logits: jnp.ndarray, labels: jnp.ndarray) -> jnp.ndarray:
  predictions = jnp.argmax(logits, axis=-1)
  return jnp.mean(jnp.argmax(labels, axis=-1) == predictions)


def model_accuracy(
    model: Callable[[jnp.ndarray], jnp.ndarray],
    batch_iterable: Iterable[Mapping[str, jnp.ndarray]]) -> jnp.ndarray:
  """Returns the accuracy of a model on a batched dataset."""
  accuracy_sum = dataset_size = 0
  for batch in batch_iterable:
    # Take batch size into account in case there is a smaller remainder batch.
    batch_size = batch['image'].shape[0]
    logits = model(batch['image'])
    accuracy_sum += accuracy(logits, batch['label']) * batch_size
    dataset_size += batch_size

  return accuracy_sum / dataset_size


# Optax is agnostic to which (if any) neural network library is used. Below we
# provide a Haiku version.
def _make_model(
    layer_dims: Sequence[int]) -> Tuple[Callable[..., Any], Callable[..., Any]]:
  """Simple multi-layer perceptron model for image classification."""

  @hk.transform
  def mlp_model(inputs: jnp.ndarray) -> jnp.ndarray:
    flattened = hk.Flatten()(inputs)
    return hk.nets.MLP(layer_dims)(flattened)

  return hk.without_apply_rng(mlp_model)


def main(unused_argv) -> None:
  train_dataset = datasets.load_image_dataset('mnist', FLAGS.batch_size)
  test_dataset = datasets.load_image_dataset('mnist', FLAGS.batch_size,
                                             datasets.Split.TEST)
  num_classes = train_dataset.element_spec['label'].shape[1]

  init_params_fn, apply_params_fn = _make_model(
      (*FLAGS.hidden_dims, num_classes))

  # Set up the fast optimizer (adam) and wrap lookahead around it.
  fast_optimizer = optax.adam(FLAGS.learning_rate)
  optimizer = optax.lookahead(fast_optimizer, FLAGS.sync_period,
                              FLAGS.slow_step_size)

  def get_loss(fast_params, batch):
    logits = apply_params_fn(fast_params, batch['image'])
    return categorical_crossentropy(logits, batch['label'])

  @jax.jit
  def train_step(params, optimizer_state, batch):
    grads = jax.grad(get_loss)(params.fast, batch)
    updates, opt_state = optimizer.update(grads, optimizer_state, params)
    return optax.apply_updates(params, updates), opt_state

  example_input = next(train_dataset.as_numpy_iterator())['image']
  initial_params = init_params_fn(random.PRNGKey(FLAGS.seed), example_input)

  # The lookahead optimizer wrapper keeps a pair of slow and fast parameters. To
  # initialize them, we create a pair of synchronized parameters from the
  # initial model parameters. The first line below is only necessary for the
  # lookahead wrapper; without it the initial parameters could be used in the
  # initialization function of the optimizer directly.
  params = optax.LookaheadParams.init_synced(initial_params)
  opt_state = optimizer.init(params)

  # Training loop
  for epoch in range(FLAGS.n_epochs):
    for batch in train_dataset.as_numpy_iterator():
      params, opt_state = train_step(params, opt_state, batch)

    # Validation is done on the slow lookahead parameters.
    eval_model = functools.partial(apply_params_fn, params.slow)
    test_acc = model_accuracy(eval_model, test_dataset.as_numpy_iterator())
    print(f'Epoch {epoch+1}: test acc: {test_acc:.2f}')


if __name__ == '__main__':
  app.run(main)



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `second_order.py`."""

import collections
import functools
import itertools

from absl.testing import absltest

import chex
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np

from optax._src import second_order


NUM_CLASSES = 2
NUM_SAMPLES = 3
NUM_FEATURES = 4


class SecondOrderTest(chex.TestCase):

  def setUp(self):
    super().setUp()

    self.data = np.random.rand(NUM_SAMPLES, NUM_FEATURES)
    self.labels = np.random.randint(NUM_CLASSES, size=NUM_SAMPLES)

    def net_fn(z):
      mlp = hk.Sequential(
          [hk.Linear(10), jax.nn.relu, hk.Linear(NUM_CLASSES)], name='mlp')
      return jax.nn.log_softmax(mlp(z))

    net = hk.without_apply_rng(hk.transform(net_fn, apply_rng=True))
    self.parameters = net.init(jax.random.PRNGKey(0), self.data)

    def loss(params, inputs, targets):
      log_probs = net.apply(params, inputs)
      return -jnp.mean(hk.one_hot(targets, NUM_CLASSES) * log_probs)

    self.loss_fn = loss

    def jax_hessian_diag(loss_fun, params, inputs, targets):
      """This is the 'ground-truth' obtained via the JAX library."""
      hess = jax.hessian(loss_fun)(params, inputs, targets)

      # Extracts the diagonal components.
      hess_diag = collections.defaultdict(dict)
      for k0, k1 in itertools.product(params.keys(), ['w', 'b']):
        params_shape = params[k0][k1].shape
        n_params = np.prod(params_shape)
        hess_diag[k0][k1] = jnp.diag(hess[k0][k1][k0][k1].reshape(
            n_params, n_params)).reshape(params_shape)
      for k, v in hess_diag.items():
        hess_diag[k] = v
      return second_order.ravel(hess_diag)

    self.hessian = jax_hessian_diag(
        self.loss_fn, self.parameters, self.data, self.labels)

  @chex.all_variants()
  def test_hessian_diag(self):
    hessian_diag_fn = self.variant(
        functools.partial(second_order.hessian_diag, self.loss_fn))
    actual = hessian_diag_fn(self.parameters, self.data, self.labels)
    np.testing.assert_array_almost_equal(self.hessian, actual, 5)

  @chex.all_variants()
  def test_fisher_diag_shape(self):
    fisher_diag_fn = self.variant(
        functools.partial(second_order.fisher_diag, self.loss_fn))
    fisher_diagonal = fisher_diag_fn(self.parameters, self.data, self.labels)
    chex.assert_equal_shape([fisher_diagonal, self.hessian])


if __name__ == '__main__':
  absltest.main()



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `schedule.py`."""

import functools

from absl.testing import absltest
from absl.testing import parameterized

import chex
import jax
import jax.numpy as jnp
import numpy as np

from optax._src import schedule
from optax._src import transform
from optax._src import wrappers


class ConstantTest(chex.TestCase):

  @chex.all_variants()
  def test_constant(self):
    """Check constant schedule."""
    # Get schedule function.
    const_value = 10
    num_steps = 15
    schedule_fn = self.variant(schedule.constant_schedule(const_value))
    # Test that generated values equal the expected schedule values.
    generated_vals = []
    for count in range(num_steps):
      # Compute next value.
      generated_vals.append(schedule_fn(count))
    # Test output.
    expected_vals = np.array([const_value] * num_steps, dtype=np.float32)
    np.testing.assert_allclose(
        expected_vals, np.array(generated_vals), atol=1e-3)


class PolynomialTest(chex.TestCase):

  @chex.all_variants()
  def test_linear(self):
    """Check linear schedule."""
    # Get schedule function.
    schedule_fn = self.variant(
        schedule.polynomial_schedule(
            init_value=10., end_value=20., power=1, transition_steps=10))
    # Test that generated values equal the expected schedule values.
    generated_vals = []
    for count in range(15):
      # Compute next value.
      generated_vals.append(schedule_fn(count))
    # Test output.
    expected_vals = np.array(list(range(10, 20)) + [20] * 5, dtype=np.float32)
    np.testing.assert_allclose(
        expected_vals, np.array(generated_vals), atol=1e-3)

  @chex.all_variants()
  def test_zero_steps_schedule(self):
    # Get schedule function.
    initial_value = 10.
    end_value = 20.

    for num_steps in [-1, 0]:
      schedule_fn = self.variant(
          schedule.polynomial_schedule(
              init_value=initial_value, end_value=end_value,
              power=1, transition_steps=num_steps))
      for count in range(15):
        np.testing.assert_allclose(schedule_fn(count), initial_value)

  @chex.all_variants()
  def test_nonlinear(self):
    """Check non-linear (quadratic) schedule."""
    # Get schedule function.
    schedule_fn = self.variant(
        schedule.polynomial_schedule(
            init_value=25., end_value=10., power=2, transition_steps=10))
    # Test that generated values equal the expected schedule values.
    generated_vals = []
    for count in range(15):
      # Compute next value.
      generated_vals.append(schedule_fn(count))
    # Test output.
    expected_vals = np.array(
        [10. + 15. * (1. - n / 10)**2 for n in range(10)] + [10] * 5,
        dtype=np.float32)
    np.testing.assert_allclose(
        expected_vals, np.array(generated_vals), atol=1e-3)

  @chex.all_variants()
  def test_with_decay_begin(self):
    """Check quadratic schedule with non-zero schedule begin."""
    # Get schedule function.
    schedule_fn = self.variant(
        schedule.polynomial_schedule(
            init_value=30., end_value=10., power=2,
            transition_steps=10, transition_begin=4))
    # Test that generated values equal the expected schedule values.
    generated_vals = []
    for count in range(20):
      # Compute next value.
      generated_vals.append(schedule_fn(count))
    # Test output.
    expected_vals = np.array(
        [30.] * 4 + [10. + 20. * (1. - n / 10)**2 for n in range(10)] +
        [10] * 6,
        dtype=np.float32)
    np.testing.assert_allclose(
        expected_vals, np.array(generated_vals), atol=1e-3)


class PiecewiseConstantTest(chex.TestCase):

  @chex.all_variants()
  def test_positive(self):
    """Check piecewise constant schedule of positive values."""
    # Get schedule function.
    schedule_fn = self.variant(
        schedule.piecewise_constant_schedule(0.1, {3: 2., 6: 0.5}))
    # Test that generated values equal the expected schedule values.
    generated_vals = []
    for count in range(10):
      # Compute next value.
      generated_vals.append(schedule_fn(count))
    # Test output.
    expected_vals = np.array([0.1, 0.1, 0.1, 0.2, 0.2, 0.2, 0.1, 0.1, 0.1, 0.1])
    np.testing.assert_allclose(
        expected_vals, np.array(generated_vals), atol=1e-3)

  @chex.all_variants()
  def test_negative(self):
    """Check piecewise constant schedule of negative values."""
    # Get schedule function.
    schedule_fn = self.variant(
        schedule.piecewise_constant_schedule(-0.1, {3: 2., 6: 0.5}))
    # Test that generated values equal the expected schedule values.
    generated_vals = []
    for count in range(10):
      # Compute next value.
      generated_vals.append(schedule_fn(count))
    # Test output.
    expected_vals = -1 * np.array(
        [0.1, 0.1, 0.1, 0.2, 0.2, 0.2, 0.1, 0.1, 0.1, 0.1])
    np.testing.assert_allclose(
        expected_vals, np.array(generated_vals), atol=1e-3)


class ExponentialTest(chex.TestCase):

  @chex.all_variants()
  @parameterized.parameters(False, True)
  def test_constant_schedule(self, staircase):
    """Checks constant schedule for exponential decay schedule."""
    num_steps = 15
    # Get schedule function.
    init_value = 1.
    schedule_fn = self.variant(
        schedule.exponential_decay(
            init_value=init_value, transition_steps=num_steps,
            decay_rate=1., staircase=staircase))
    # Test that generated values equal the expected schedule values.
    generated_vals = []
    for count in range(num_steps):
      generated_vals.append(schedule_fn(count))
    expected_vals = np.array([init_value] * num_steps, dtype=np.float32)
    np.testing.assert_allclose(
        expected_vals, np.array(generated_vals), atol=1e-3)

  @chex.all_variants()
  @parameterized.parameters(False, True)
  def test_nonvalid_transition_steps(self, staircase):
    """Checks nonvalid decay steps results in a constant schedule."""
    init_value = 1.
    for transition_steps in [-1, 0]:
      schedule_fn = self.variant(
          schedule.exponential_decay(
              init_value=init_value, transition_steps=transition_steps,
              decay_rate=1., staircase=staircase))
      for count in range(15):
        np.testing.assert_allclose(schedule_fn(count), init_value)

  @chex.all_variants()
  @parameterized.parameters(False, True)
  def test_nonvalid_decay_rate(self, staircase):
    """Checks nonvalid decay steps results in a constant schedule."""
    init_value = 1.
    schedule_fn = self.variant(
        schedule.exponential_decay(
            init_value=init_value, transition_steps=2,
            decay_rate=0., staircase=staircase))
    for count in range(15):
      np.testing.assert_allclose(schedule_fn(count), init_value)

  @chex.all_variants()
  @parameterized.parameters((False, 0), (True, 0), (False, 5), (True, 5))
  def test_exponential(self, staircase, transition_begin):
    """Checks non-linear (quadratic) schedule."""
    # Get schedule function.
    init_value = 1.
    num_steps = 15
    transition_steps = 2
    decay_rate = 2.
    schedule_fn = self.variant(
        schedule.exponential_decay(
            init_value=init_value, transition_steps=transition_steps,
            decay_rate=decay_rate, transition_begin=transition_begin,
            staircase=staircase))

    # Test that generated values equal the expected schedule values.
    def _staircased(count):
      p = count / transition_steps
      if staircase:
        p = np.floor(p)
      return p

    generated_vals = []
    for count in range(num_steps + transition_begin):
      generated_vals.append(schedule_fn(count))
    expected_vals = np.array(
        [init_value] * transition_begin + [
            init_value * np.power(decay_rate, _staircased(count))
            for count in range(num_steps)
        ],
        dtype=np.float32)
    np.testing.assert_allclose(
        expected_vals, np.array(generated_vals), atol=1e-3)

  @chex.all_variants()
  @parameterized.parameters(
      (0.2, 0.1, False), (1.0, 0.1, False), (2.0, 3.0, False),
      (0.2, 0.1, True), (1.0, 0.1, True), (2.0, 3.0, True))
  def test_end_value_with_staircase(self, decay_rate, end_value, staircase):
    # Get schedule function.
    init_value = 1.
    num_steps = 11
    transition_steps = 2
    transition_begin = 3
    schedule_fn = self.variant(
        schedule.exponential_decay(
            init_value=init_value, transition_steps=transition_steps,
            decay_rate=decay_rate, transition_begin=transition_begin,
            staircase=staircase, end_value=end_value))

    # Test that generated values equal the expected schedule values.
    def _staircased(count):
      p = count / transition_steps
      if staircase:
        p = np.floor(p)
      return p

    generated_vals = []
    for count in range(num_steps + transition_begin):
      generated_vals.append(schedule_fn(count))
    expected_vals = np.array(
        [init_value] * transition_begin + [
            init_value * np.power(decay_rate, _staircased(count))
            for count in range(num_steps)
        ],
        dtype=np.float32)

    if decay_rate < 1.0:
      expected_vals = np.maximum(expected_vals, end_value)
    else:
      expected_vals = np.minimum(expected_vals, end_value)

    np.testing.assert_allclose(
        expected_vals, np.array(generated_vals), atol=1e-3)


class CosineDecayTest(chex.TestCase):

  @chex.all_variants()
  def test_decay_count_smaller_count(self):
    """Check cosine schedule decay for the entire training schedule."""
    initial_value = 0.1
    schedule_fn = self.variant(
        schedule.cosine_decay_schedule(initial_value, 10, 0.0))
    # Test that generated values equal the expected schedule values.
    generated_vals = []
    for count in range(10):
      # Compute next value.
      generated_vals.append(schedule_fn(count))
    # Test output.
    expected_multipliers = np.array(
        0.5 + 0.5 * np.cos(
            np.pi * np.array(
                [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9])))
    np.testing.assert_allclose(
        initial_value * expected_multipliers,
        np.array(generated_vals), atol=1e-3)

  @chex.all_variants()
  def test_decay_count_greater_count(self):
    """Check cosine schedule decay for a part of the training schedule."""
    initial_value = 0.1
    schedule_fn = self.variant(
        schedule.cosine_decay_schedule(initial_value, 5, 0.0))
    # Test that generated values equal the expected schedule values.
    generated_vals = []
    for count in range(12):
      # Compute next value.
      generated_vals.append(schedule_fn(count))

    # Test output.
    expected_multipliers = np.array(
        0.5 + 0.5 * np.cos(
            np.pi * np.array(
                [0.0, 0.2, 0.4, 0.6, 0.8, 1., 1., 1., 1., 1., 1., 1.])))
    np.testing.assert_allclose(
        initial_value * expected_multipliers,
        np.array(generated_vals), atol=1e-3)

  @chex.all_variants()
  def test_decay_count_greater_count_with_alpha(self):
    """Check cosine schedule decay for a part of the training schedule."""
    # Get schedule function.
    initial_value = 0.1
    schedule_fn = self.variant(
        schedule.cosine_decay_schedule(initial_value, 5, 0.1))
    # Test that generated values equal the expected schedule values.
    generated_vals = []
    for count in range(12):
      # Compute next value.
      generated_vals.append(schedule_fn(count))

    # Test output.
    expected_multipliers = np.array(
        0.5 + 0.5 * np.cos(
            np.pi * np.array(
                [0.0, 0.2, 0.4, 0.6, 0.8, 1., 1., 1., 1., 1., 1., 1.])))
    expected_multipliers = 0.9 * expected_multipliers + 0.1
    np.testing.assert_allclose(
        initial_value * expected_multipliers,
        np.array(generated_vals), atol=1e-3)


class PiecewiseInterpolateTest(chex.TestCase):

  @chex.all_variants()
  def test_linear_piecewise(self):
    schedule_fn = self.variant(schedule.piecewise_interpolate_schedule(
        'linear', 200., {5: 1.5, 10: 0.25}))
    generated_vals = [schedule_fn(step) for step in range(13)]
    expected_vals = [200., 220., 240., 260., 280., 300., 255., 210., 165.,
                     120., 75., 75., 75.]
    np.testing.assert_allclose(generated_vals, expected_vals, atol=1e-3)

  @chex.all_variants()
  def test_cos_piecewise(self):
    schedule_fn = self.variant(schedule.piecewise_interpolate_schedule(
        'cosine', 400., {5: 1.2, 3: 0.6, 7: 1.}))
    generated_vals = [schedule_fn(step) for step in range(9)]
    expected_vals = [400., 360., 280., 240., 264., 288., 288., 288., 288.]
    np.testing.assert_allclose(generated_vals, expected_vals, atol=1e-3)

  @chex.all_variants()
  def test_empty_dict(self):
    schedule_fn = self.variant(schedule.piecewise_interpolate_schedule(
        'linear', 13., {}))
    generated_vals = [schedule_fn(step) for step in range(5)]
    expected_vals = [13., 13., 13., 13., 13.]
    np.testing.assert_allclose(generated_vals, expected_vals, atol=1e-3)

  @chex.all_variants()
  def test_no_dict(self):
    schedule_fn = self.variant(schedule.piecewise_interpolate_schedule(
        'cosine', 17.))
    generated_vals = [schedule_fn(step) for step in range(3)]
    expected_vals = [17., 17., 17.]
    np.testing.assert_allclose(generated_vals, expected_vals, atol=1e-3)

  def test_invalid_type(self):
    # pytype: disable=wrong-arg-types
    with self.assertRaises(ValueError):
      schedule.piecewise_interpolate_schedule('linar', 13.)
    with self.assertRaises(ValueError):
      schedule.piecewise_interpolate_schedule('', 13., {5: 3.})
    with self.assertRaises(ValueError):
      schedule.piecewise_interpolate_schedule(None, 13., {})
    # pytype: enable=wrong-arg-types

  def test_invalid_scale(self):
    with self.assertRaises(ValueError):
      schedule.piecewise_interpolate_schedule('linear', 13., {5: -3})


class OneCycleTest(chex.TestCase):

  @chex.all_variants()
  def test_linear(self):
    schedule_fn = self.variant(schedule.linear_onecycle_schedule(
        transition_steps=10,
        peak_value=1000,
        pct_start=0.3,
        pct_final=0.7,
        div_factor=10.,
        final_div_factor=100.))

    generated_vals = [schedule_fn(step) for step in range(12)]
    expected_vals = [100., 400., 700., 1000., 775., 550., 325., 100., 67.,
                     34., 1., 1.]
    np.testing.assert_allclose(generated_vals, expected_vals, atol=1e-3)

  @chex.all_variants()
  def test_cosine(self):
    schedule_fn = self.variant(schedule.cosine_onecycle_schedule(
        transition_steps=5,
        peak_value=1000.,
        pct_start=0.4,
        div_factor=10.,
        final_div_factor=100.))

    generated_vals = [schedule_fn(step) for step in range(7)]
    expected_vals = [100., 550., 1000., 750.25, 250.75, 1., 1.]
    np.testing.assert_allclose(generated_vals, expected_vals, atol=1e-3)

  def test_nonpositive_transition_steps(self):
    with self.assertRaises(ValueError):
      schedule.cosine_onecycle_schedule(transition_steps=0, peak_value=5.)
    with self.assertRaises(ValueError):
      schedule.linear_onecycle_schedule(transition_steps=0, peak_value=5.)


class InjectHyperparamsTest(chex.TestCase):
  """Tests for the inject_hyperparams wrapper."""

  @chex.all_variants
  def test_updates(self):
    optim = schedule.inject_hyperparams(transform.scale)(  # stateless
        step_size=schedule.piecewise_constant_schedule(
            3.0, {2: 5, 8: 2, 13: 1.5}))

    params = [jnp.zeros([], dtype=jnp.float32)]
    state = self.variant(optim.init)(params)
    update_fn = self.variant(optim.update)
    expected_step_size = [3.0]*2 + [15.0]*6 + [30.0]*5 + [45.0]*3

    grads = [jnp.ones([], dtype=jnp.float32)]
    for i in range(15):
      updates, state = update_fn(grads, state, params=params)
      np.testing.assert_almost_equal(updates[0], expected_step_size[i+1])

  @chex.all_variants
  def test_hyperparams_state(self):
    optim = schedule.inject_hyperparams(transform.trace)(  # stateful
        decay=schedule.piecewise_constant_schedule(
            0.8, {4: 0.5, 10: 1.25}),
        nesterov=True)

    params = [jnp.zeros([2, 3]) for _ in range(3)]
    state = self.variant(optim.init)(params)
    update_fn = self.variant(optim.update)

    expected_mom = [0.8]*4 + [0.4]*6 + [0.5]*2
    grads = jax.tree_map(jnp.ones_like, params)
    for i in range(12):
      np.testing.assert_almost_equal(state.hyperparams['decay'],
                                     expected_mom[i])
      _, state = update_fn(grads, state)

    np.testing.assert_almost_equal(state.hyperparams['decay'],
                                   expected_mom[-1])

  @chex.all_variants
  def test_constant_hyperparams(self):
    optim = schedule.inject_hyperparams(transform.scale_by_adam)(b1=0., b2=0.)

    params = [jnp.zeros([2, 3]) for _ in range(3)]
    state = self.variant(optim.init)(params)
    update_fn = self.variant(optim.update)

    grads = jax.tree_map(jnp.ones_like, params)
    for _ in range(5):
      updates, state = update_fn(grads, state, params)
      np.testing.assert_almost_equal(state.hyperparams['b1'], 0.0)
      np.testing.assert_almost_equal(state.hyperparams['b2'], 0.0)
      np.testing.assert_almost_equal(state.hyperparams['eps'], 1e-8)
      np.testing.assert_almost_equal(state.hyperparams['eps_root'], 0.0)
      assert 'eps' in state.hyperparams
      chex.assert_tree_all_close(updates, grads)

  @chex.all_variants
  def test_overriding_hyperparam(self):
    optim = schedule.inject_hyperparams(transform.clip_by_global_norm)(0.1)
    params = jnp.zeros((3, 5, 7))
    state = self.variant(optim.init)(params)
    update_fn = self.variant(optim.update)

    grads = jnp.ones_like(params)
    for i in range(5):
      state.hyperparams['max_norm'] = i
      updates, state = update_fn(grads, state)
      assert np.isclose(jnp.linalg.norm(updates.ravel()), i)

  @chex.all_variants
  @parameterized.named_parameters(('string', 'mask'), ('list', ['mask']))
  def test_static_args(self, static_args):
    @functools.partial(schedule.inject_hyperparams, static_args=static_args)
    def custom_optim(learning_rate, mask):
      return wrappers.masked(transform.scale(-learning_rate), mask)

    optim = custom_optim(
        0.1, functools.partial(jax.tree_map, lambda x: x.ndim > 1))
    params = [jnp.ones((1, 2)), jnp.ones(2), jnp.ones((1, 1, 1))]
    grads = params
    state = self.variant(optim.init)(params)
    updates, state = self.variant(optim.update)(grads, state)
    expected_updates = jax.tree_map(lambda x: -0.1 * x if x.ndim > 1 else x,
                                    grads)

    assert set(state.hyperparams.keys()) == {'learning_rate'}, state.hyperparams
    chex.assert_tree_all_close(updates, expected_updates)

  @chex.all_variants
  @parameterized.named_parameters(('one_arg', 'b1'), ('two_arg', ['b1', 'b2']))
  def test_numeric_static_args(self, static_args):
    optim = schedule.inject_hyperparams(
        transform.scale_by_adam, static_args=static_args)(b1=0.9, b2=0.95)

    params = [jnp.ones((1, 2)), jnp.ones(2), jnp.ones((1, 1, 1))]
    grads = params
    state = self.variant(optim.init)(params)
    _, state = self.variant(optim.update)(grads, state)

    assert not set(state.hyperparams.keys()).intersection(set(static_args))

  @parameterized.named_parameters(('string', 'lr'), ('list', ['lr']))
  def test_static_args_error(self, static_args):
    with self.assertRaises(ValueError):
      schedule.inject_hyperparams(transform.scale, static_args=static_args)


if __name__ == '__main__':
  absltest.main()



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flexibly compose gradient transformations."""

from optax._src import transform
GradientTransformation = transform.GradientTransformation


def chain(*args: GradientTransformation) -> GradientTransformation:
  """Applies a list of chainable update transformations.

  Given a sequence of chainable transforms, `chain` returns an `init_fn`
  that constructs a `state` by concatenating the states of the individual
  transforms, and returns an `update_fn` which chains the update transformations
  feeding the appropriate state to each.

  Args:
    *args: a sequence of chainable (init_fn, update_fn) tuples.

  Returns:
    A single (init_fn, update_fn) tuple.
  """

  init_fns, update_fns = zip(*args)

  def init_fn(params):
    return [fn(params) for fn in init_fns]

  def update_fn(updates, state, params=None):
    if len(update_fns) != len(state):
      raise ValueError('The number of updates and states has to be the same in '
                       'chain! Make sure you have called init first!')

    new_state = []
    for s, fn in zip(state, update_fns):
      updates, new_s = fn(updates, s, params)
      new_state.append(new_s)
    return updates, new_state

  return GradientTransformation(init_fn, update_fn)



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `stochastic_gradient_estimators.py`."""

from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import jax.numpy as jnp
import numpy as np

from optax._src import stochastic_gradient_estimators as sge
from optax._src import utils


# Set seed for deterministic sampling.
np.random.seed(42)


_estimator_to_num_samples = {
    sge.score_function_jacobians: 5 * 10**5,
    sge.measure_valued_jacobians: 10**5,
    sge.pathwise_jacobians: 5 * 10**4,
}

_weighted_estimator_to_num_samples = {
    sge.score_function_jacobians: 5 * 10**6,
    sge.measure_valued_jacobians: 5 * 10**5,
    sge.pathwise_jacobians: 5 * 10**4,
}


def _ones(dims):
  return jnp.ones(shape=(dims), dtype=jnp.float32)


def _assert_equal(actual, expected, rtol=1e-2, atol=1e-2):
  """Asserts that arrays are equal."""
  # Note: assert_allclose does not check shapes
  chex.assert_equal_shape((actual, expected))

  # We get around the bug https://github.com/numpy/numpy/issues/13801
  zero_indices = np.argwhere(expected == 0)
  if not np.all(np.abs(actual[zero_indices]) <= atol):
    raise AssertionError('Larger than {} diff in {}'.format(
        atol, actual[zero_indices]))

  non_zero_indices = np.argwhere(expected != 0)
  np.testing.assert_allclose(
      np.asarray(actual)[non_zero_indices],
      expected[non_zero_indices], rtol, atol)


def _estimator_variant(variant, estimator):
  return variant(estimator, static_argnums=(0, 2, 4))


def _measure_valued_variant(variant):
  return variant(
      sge.measure_valued_jacobians,
      static_argnums=(0, 2, 4, 5))


class GradientEstimatorsTest(chex.TestCase):

  @chex.all_variants
  @parameterized.named_parameters(
      chex.params_product([
          ('_score_function_jacobians', sge.score_function_jacobians),
          ('_pathwise_jacobians', sge.pathwise_jacobians),
          ('_measure_valued_jacobians', sge.measure_valued_jacobians),
      ], [
          ('0.1', 0.1),
          ('0.5', 0.5),
          ('0.9', 0.9),
      ],
                          named=True))
  def testConstantFunction(self, estimator, constant):
    data_dims = 3
    num_samples = _estimator_to_num_samples[estimator]

    effective_mean = 1.5
    mean = effective_mean * _ones(data_dims)

    effective_log_scale = 0.0
    log_scale = effective_log_scale * _ones(data_dims)
    rng = jax.random.PRNGKey(1)

    jacobians = _estimator_variant(self.variant, estimator)(
        lambda x: jnp.array(constant), [mean, log_scale],
        utils.multi_normal, rng, num_samples)

    # Average over the number of samples.
    mean_jacobians = jacobians[0]
    chex.assert_shape(mean_jacobians, (num_samples, data_dims))
    mean_grads = np.mean(mean_jacobians, axis=0)
    expected_mean_grads = np.zeros(data_dims, dtype=np.float32)

    log_scale_jacobians = jacobians[1]
    chex.assert_shape(log_scale_jacobians, (num_samples, data_dims))
    log_scale_grads = np.mean(log_scale_jacobians, axis=0)
    expected_log_scale_grads = np.zeros(data_dims, dtype=np.float32)

    _assert_equal(mean_grads, expected_mean_grads, atol=5e-3)
    _assert_equal(log_scale_grads, expected_log_scale_grads, atol=5e-3)

  @chex.all_variants
  @parameterized.named_parameters(
      chex.params_product([
          ('_score_function_jacobians', sge.score_function_jacobians),
          ('_pathwise_jacobians', sge.pathwise_jacobians),
          ('_measure_valued_jacobians', sge.measure_valued_jacobians),
      ], [
          ('0.5_-1.', 0.5, -1.),
          ('0.7_0.0)', 0.7, 0.0),
          ('0.8_0.1', 0.8, 0.1),
      ],
                          named=True))
  def testLinearFunction(self, estimator, effective_mean, effective_log_scale):
    data_dims = 3
    num_samples = _estimator_to_num_samples[estimator]
    rng = jax.random.PRNGKey(1)

    mean = effective_mean * _ones(data_dims)
    log_scale = effective_log_scale * _ones(data_dims)

    jacobians = _estimator_variant(self.variant, estimator)(
        np.sum, [mean, log_scale],
        utils.multi_normal, rng, num_samples)

    mean_jacobians = jacobians[0]
    chex.assert_shape(mean_jacobians, (num_samples, data_dims))
    mean_grads = np.mean(mean_jacobians, axis=0)
    expected_mean_grads = np.ones(data_dims, dtype=np.float32)

    log_scale_jacobians = jacobians[1]
    chex.assert_shape(log_scale_jacobians, (num_samples, data_dims))
    log_scale_grads = np.mean(log_scale_jacobians, axis=0)
    expected_log_scale_grads = np.zeros(data_dims, dtype=np.float32)

    _assert_equal(mean_grads, expected_mean_grads)
    _assert_equal(log_scale_grads, expected_log_scale_grads)

  @chex.all_variants
  @parameterized.named_parameters(
      chex.params_product([
          ('_score_function_jacobians', sge.score_function_jacobians),
          ('_pathwise_jacobians', sge.pathwise_jacobians),
          ('_measure_valued_jacobians', sge.measure_valued_jacobians),
      ], [
          ('1.0_0.3', 1.0, 0.3),
      ],
                          named=True))
  def testQuadraticFunction(
      self, estimator, effective_mean, effective_log_scale):
    data_dims = 3
    num_samples = _estimator_to_num_samples[estimator]
    rng = jax.random.PRNGKey(1)

    mean = effective_mean * _ones(data_dims)
    log_scale = effective_log_scale * _ones(data_dims)

    jacobians = _estimator_variant(self.variant, estimator)(
        lambda x: np.sum(x**2) / 2, [mean, log_scale],
        utils.multi_normal, rng, num_samples)

    mean_jacobians = jacobians[0]
    chex.assert_shape(mean_jacobians, (num_samples, data_dims))
    mean_grads = np.mean(mean_jacobians, axis=0)
    expected_mean_grads = effective_mean * np.ones(
        data_dims, dtype=np.float32)

    log_scale_jacobians = jacobians[1]
    chex.assert_shape(log_scale_jacobians, (num_samples, data_dims))
    log_scale_grads = np.mean(log_scale_jacobians, axis=0)
    expected_log_scale_grads = np.exp(2 * effective_log_scale) * np.ones(
        data_dims, dtype=np.float32)

    _assert_equal(mean_grads, expected_mean_grads, atol=5e-2)
    _assert_equal(log_scale_grads, expected_log_scale_grads, atol=5e-2)

  @chex.all_variants
  @parameterized.named_parameters(
      chex.params_product([
          ('_score_function_jacobians', sge.score_function_jacobians),
          ('_pathwise_jacobians', sge.pathwise_jacobians),
          ('_measure_valued_jacobians', sge.measure_valued_jacobians),
      ], [
          ('case_1', [1.0, 2.0, 3.], [-1., 0.3, -2.], [1., 1., 1.]),
          ('case_2', [1.0, 2.0, 3.], [-1., 0.3, -2.], [4., 2., 3.]),
          ('case_3', [1.0, 2.0, 3.], [0.1, 0.2, 0.1], [10., 5., 1.]),
      ],
                          named=True))
  def testWeightedLinear(
      self, estimator, effective_mean, effective_log_scale, weights):
    num_samples = _weighted_estimator_to_num_samples[estimator]
    rng = jax.random.PRNGKey(1)

    mean = jnp.array(effective_mean)
    log_scale = jnp.array(effective_log_scale)
    weights = jnp.array(weights)

    data_dims = len(effective_mean)

    function = lambda x: jnp.sum(x * weights)
    jacobians = _estimator_variant(self.variant, estimator)(
        function, [mean, log_scale],
        utils.multi_normal, rng, num_samples)

    mean_jacobians = jacobians[0]
    chex.assert_shape(mean_jacobians, (num_samples, data_dims))
    mean_grads = np.mean(mean_jacobians, axis=0)

    log_scale_jacobians = jacobians[1]
    chex.assert_shape(log_scale_jacobians, (num_samples, data_dims))
    log_scale_grads = np.mean(log_scale_jacobians, axis=0)

    expected_mean_grads = weights
    expected_log_scale_grads = np.zeros(data_dims, dtype=np.float32)

    _assert_equal(mean_grads, expected_mean_grads, atol=5e-2)
    _assert_equal(log_scale_grads, expected_log_scale_grads, atol=5e-2)

  @chex.all_variants
  @parameterized.named_parameters(
      chex.params_product([
          ('_score_function_jacobians', sge.score_function_jacobians),
          ('_pathwise_jacobians', sge.pathwise_jacobians),
          ('_measure_valued_jacobians', sge.measure_valued_jacobians),
      ], [
          ('case_1', [1.0, 2.0, 3.], [-1., 0.3, -2.], [1., 1., 1.]),
          ('case_2', [1.0, 2.0, 3.], [-1., 0.3, -2.], [4., 2., 3.]),
          ('case_3', [1.0, 2.0, 3.], [0.1, 0.2, 0.1], [3., 5., 1.]),
      ],
                          named=True))
  def testWeightedQuadratic(
      self, estimator, effective_mean, effective_log_scale, weights):
    num_samples = _weighted_estimator_to_num_samples[estimator]
    rng = jax.random.PRNGKey(1)

    mean = jnp.array(effective_mean, dtype=jnp.float32)
    log_scale = jnp.array(effective_log_scale, dtype=jnp.float32)
    weights = jnp.array(weights, dtype=jnp.float32)

    data_dims = len(effective_mean)

    function = lambda x: jnp.sum(x * weights) ** 2
    jacobians = _estimator_variant(self.variant, estimator)(
        function, [mean, log_scale],
        utils.multi_normal, rng, num_samples)

    mean_jacobians = jacobians[0]
    chex.assert_shape(mean_jacobians, (num_samples, data_dims))
    mean_grads = np.mean(mean_jacobians, axis=0)

    log_scale_jacobians = jacobians[1]
    chex.assert_shape(log_scale_jacobians, (num_samples, data_dims))
    log_scale_grads = np.mean(log_scale_jacobians, axis=0)

    expected_mean_grads = 2 * weights * np.sum(weights * mean)
    effective_scale = np.exp(log_scale)
    expected_scale_grads = 2 * weights ** 2 * effective_scale
    expected_log_scale_grads = expected_scale_grads * effective_scale

    _assert_equal(mean_grads, expected_mean_grads, atol=1e-1, rtol=1e-1)
    _assert_equal(
        log_scale_grads, expected_log_scale_grads, atol=1e-1, rtol=1e-1)

  @chex.all_variants
  @parameterized.named_parameters(
      chex.params_product(
          [
              ('_sum_cos_x', [1.0], [1.0], lambda x: jnp.sum(jnp.cos(x))),
              # Need to ensure that the mean is not too close to 0.
              ('_sum_log_x', [10.0], [0.0], lambda x: jnp.sum(jnp.log(x))),
              ('_sum_cos_2x', [1.0, 2.0], [1.0, -2
                                          ], lambda x: jnp.sum(jnp.cos(2 * x))),
              ('_cos_sum_2x', [1.0, 2.0], [1.0, -2
                                          ], lambda x: jnp.cos(jnp.sum(2 * x))),
          ],
          [
              ('coupling', True),
              ('nocoupling', False),
          ],
          named=True))
  def testNonPolynomialFunctionConsistencyWithPathwise(self, effective_mean,
                                                       effective_log_scale,
                                                       function, coupling):
    num_samples = 10**5
    rng = jax.random.PRNGKey(1)
    measure_rng, pathwise_rng = jax.random.split(rng)

    mean = jnp.array(effective_mean, dtype=jnp.float32)
    log_scale = jnp.array(effective_log_scale, dtype=jnp.float32)
    data_dims = len(effective_mean)

    measure_valued_jacobians = _measure_valued_variant(self.variant)(
        function, [mean, log_scale],
        utils.multi_normal, measure_rng, num_samples, coupling)

    measure_valued_mean_jacobians = measure_valued_jacobians[0]
    chex.assert_shape(measure_valued_mean_jacobians, (num_samples, data_dims))
    measure_valued_mean_grads = np.mean(measure_valued_mean_jacobians, axis=0)

    measure_valued_log_scale_jacobians = measure_valued_jacobians[1]
    chex.assert_shape(
        measure_valued_log_scale_jacobians, (num_samples, data_dims))
    measure_valued_log_scale_grads = np.mean(
        measure_valued_log_scale_jacobians, axis=0)

    pathwise_jacobians = _estimator_variant(
        self.variant, sge.pathwise_jacobians)(function, [mean, log_scale],
                                              utils.multi_normal, pathwise_rng,
                                              num_samples)

    pathwise_mean_jacobians = pathwise_jacobians[0]
    chex.assert_shape(pathwise_mean_jacobians, (num_samples, data_dims))
    pathwise_mean_grads = np.mean(pathwise_mean_jacobians, axis=0)

    pathwise_log_scale_jacobians = pathwise_jacobians[1]
    chex.assert_shape(pathwise_log_scale_jacobians, (num_samples, data_dims))
    pathwise_log_scale_grads = np.mean(pathwise_log_scale_jacobians, axis=0)

    _assert_equal(
        pathwise_mean_grads, measure_valued_mean_grads, rtol=5e-1, atol=1e-1)
    _assert_equal(
        pathwise_log_scale_grads, measure_valued_log_scale_grads,
        rtol=5e-1, atol=1e-1)


class MeasuredValuedEstimatorsTest(chex.TestCase):

  @chex.all_variants
  @parameterized.parameters([True, False])
  def testRaisesErrorForNonGaussian(self, coupling):
    num_samples = 10**5
    rng = jax.random.PRNGKey(1)

    function = lambda x: jnp.sum(x) ** 2

    mean = jnp.array(0, dtype=jnp.float32)
    log_scale = jnp.array(0., dtype=jnp.float32)

    class TestDist():

      def __init__(self, params):
        self._params = params

      def sample(self, n):
        return np.zeros(n)

    with self.assertRaises(ValueError):
      _measure_valued_variant(self.variant)(
          function, [mean, log_scale],
          TestDist, rng, num_samples, coupling)


if __name__ == '__main__':
  absltest.main()



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests of equivalence between optax and other optimiser libraries."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from flax import optim
from jax.experimental import optimizers
import jax.numpy as jnp

from optax._src import alias
from optax._src import update


STEPS = 50
LR = 1e-2
LR_SCHED = lambda _: LR  # Trivial constant "schedule".


class ExperimentalOptimizersEquivalenceTest(chex.TestCase):

  def setUp(self):
    super().setUp()
    self.init_params = (jnp.array([1., 2.]), jnp.array([3., 4.]))
    self.per_step_updates = (jnp.array([500., 5.]), jnp.array([300., 3.]))

  @chex.all_variants()
  @parameterized.named_parameters(
      ('sgd', alias.sgd(LR, 0.0),
       optimizers.sgd(LR), 1e-5),
      ('adam', alias.adam(LR, 0.9, 0.999, 1e-8),
       optimizers.adam(LR, 0.9, 0.999), 1e-4),
      ('rmsprop', alias.rmsprop(LR, decay=.9, eps=0.1),
       optimizers.rmsprop(LR, .9, 0.1), 1e-5),
      ('rmsprop_momentum', alias.rmsprop(
          LR, decay=.9, eps=0.1, momentum=0.9),
       optimizers.rmsprop_momentum(LR, .9, 0.1, 0.9), 1e-5),
      ('adagrad', alias.adagrad(LR, 0., 0.,),
       optimizers.adagrad(LR, 0.), 1e-5),
      ('sgd', alias.sgd(LR_SCHED, 0.0),
       optimizers.sgd(LR), 1e-5),
      ('adam', alias.adam(LR_SCHED, 0.9, 0.999, 1e-8),
       optimizers.adam(LR, 0.9, 0.999), 1e-4),
      ('rmsprop', alias.rmsprop(LR_SCHED, decay=.9, eps=0.1),
       optimizers.rmsprop(LR, .9, 0.1), 1e-5),
      ('rmsprop_momentum', alias.rmsprop(
          LR_SCHED, decay=.9, eps=0.1, momentum=0.9),
       optimizers.rmsprop_momentum(LR, .9, 0.1, 0.9), 1e-5),
      ('adagrad', alias.adagrad(LR_SCHED, 0., 0.,),
       optimizers.adagrad(LR, 0.), 1e-5),
  )
  def test_jax_optimizer_equivalent(self, optax_optimizer, jax_optimizer, rtol):

    # experimental/optimizers.py
    jax_params = self.init_params
    opt_init, opt_update, get_params = jax_optimizer
    state = opt_init(jax_params)
    for i in range(STEPS):
      state = opt_update(i, self.per_step_updates, state)
      jax_params = get_params(state)

    # optax
    optax_params = self.init_params
    state = optax_optimizer.init(optax_params)

    @self.variant
    def step(updates, state):
      return optax_optimizer.update(updates, state)

    for _ in range(STEPS):
      updates, state = step(self.per_step_updates, state)
      optax_params = update.apply_updates(optax_params, updates)

    # Check equivalence.
    chex.assert_tree_all_close(jax_params, optax_params, rtol=rtol)


class FlaxOptimizersEquivalenceTest(chex.TestCase):

  def setUp(self):
    super().setUp()
    self.init_params = (jnp.array([1., 2.]), jnp.array([3., 4.]))
    self.per_step_updates = (jnp.array([500., 5.]), jnp.array([300., 3.]))

  @parameterized.named_parameters(
      ('sgd',
       alias.sgd(LR),
       optim.GradientDescent(LR)),
      ('momentum',
       alias.sgd(LR, momentum=0.9),
       optim.Momentum(LR, beta=0.9)),  # Different names.
      ('nesterov_momentum',
       alias.sgd(LR, momentum=0.9, nesterov=True),
       optim.Momentum(LR, beta=0.9, nesterov=True)),
      ('rmsprop',
       alias.rmsprop(LR),
       optim.RMSProp(LR)),
      ('centered_rmsprop',
       alias.rmsprop(LR, centered=True),
       optim.RMSProp(LR, centered=True)),
      ('adam',
       alias.adam(LR),
       optim.Adam(LR)),
      ('adam_w',
       alias.adamw(LR, weight_decay=1e-4),
       optim.Adam(LR, weight_decay=1e-4)),  # Different name.
      ('adagrad',
       alias.adagrad(LR, initial_accumulator_value=0.),  # Different default!
       optim.Adagrad(LR)),
      ('lamb',
       alias.lamb(LR),
       optim.LAMB(LR)),
  )
  def test_flax_optim_equivalence(self, optax_optimizer, flax_optimizer):

    # flax/optim
    flax_params = self.init_params
    flax_optimizer = flax_optimizer.create(flax_params)
    for _ in range(STEPS):
      flax_optimizer = flax_optimizer.apply_gradient(
          self.per_step_updates)
      flax_params = flax_optimizer.target

    # optax
    optax_params = self.init_params
    state = optax_optimizer.init(optax_params)
    for _ in range(STEPS):
      updates, state = optax_optimizer.update(
          self.per_step_updates, state, optax_params)
      optax_params = update.apply_updates(optax_params, updates)

    # Check equivalence.
    chex.assert_tree_all_close(flax_params, optax_params, rtol=1e-4)


if __name__ == '__main__':
  absltest.main()



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Optax: composable gradient processing and optimization, in JAX."""

from optax._src.alias import adabelief
from optax._src.alias import adagrad
from optax._src.alias import adam
from optax._src.alias import adamw
from optax._src.alias import dpsgd
from optax._src.alias import fromage
from optax._src.alias import lamb
from optax._src.alias import noisy_sgd
from optax._src.alias import radam
from optax._src.alias import rmsprop
from optax._src.alias import sgd
from optax._src.alias import yogi
from optax._src.combine import chain
from optax._src.control_variates import control_delta_method
from optax._src.control_variates import control_variates_jacobians
from optax._src.control_variates import moving_avg_baseline
from optax._src.linear_algebra import matrix_inverse_pth_root
from optax._src.linear_algebra import power_iteration
from optax._src.loss import cosine_distance
from optax._src.loss import huber_loss
from optax._src.loss import l2_loss
from optax._src.loss import sigmoid_binary_cross_entropy
from optax._src.loss import smooth_labels
from optax._src.loss import softmax_cross_entropy
from optax._src.privacy import differentially_private_aggregate
from optax._src.privacy import DifferentiallyPrivateAggregateState
from optax._src.schedule import constant_schedule
from optax._src.schedule import cosine_decay_schedule
from optax._src.schedule import cosine_onecycle_schedule
from optax._src.schedule import exponential_decay
from optax._src.schedule import inject_hyperparams
from optax._src.schedule import InjectHyperparamsState
from optax._src.schedule import linear_onecycle_schedule
from optax._src.schedule import piecewise_constant_schedule
from optax._src.schedule import piecewise_interpolate_schedule
from optax._src.schedule import polynomial_schedule
from optax._src.schedule import Schedule
from optax._src.second_order import fisher_diag
from optax._src.second_order import hessian_diag
from optax._src.second_order import hvp
from optax._src.stochastic_gradient_estimators import measure_valued_jacobians
from optax._src.stochastic_gradient_estimators import pathwise_jacobians
from optax._src.stochastic_gradient_estimators import score_function_jacobians
from optax._src.transform import adaptive_grad_clip
from optax._src.transform import add_decayed_weights
from optax._src.transform import add_noise
from optax._src.transform import AddDecayedWeightsState
from optax._src.transform import additive_weight_decay
from optax._src.transform import AdditiveWeightDecayState
from optax._src.transform import AddNoiseState
from optax._src.transform import apply_every
from optax._src.transform import ApplyEvery
from optax._src.transform import centralize
from optax._src.transform import clip
from optax._src.transform import clip_by_global_norm
from optax._src.transform import ClipByGlobalNormState
from optax._src.transform import ClipState
from optax._src.transform import global_norm
from optax._src.transform import GradientTransformation
from optax._src.transform import identity
from optax._src.transform import keep_params_nonnegative
from optax._src.transform import NonNegativeParamsState
from optax._src.transform import OptState
from optax._src.transform import Params
from optax._src.transform import scale
from optax._src.transform import scale_by_adam
from optax._src.transform import scale_by_belief
from optax._src.transform import scale_by_radam
from optax._src.transform import scale_by_rms
from optax._src.transform import scale_by_rss
from optax._src.transform import scale_by_schedule
from optax._src.transform import scale_by_stddev
from optax._src.transform import scale_by_trust_ratio
from optax._src.transform import scale_by_yogi
from optax._src.transform import ScaleByAdamState
from optax._src.transform import ScaleByFromageState
from optax._src.transform import ScaleByRmsState
from optax._src.transform import ScaleByRssState
from optax._src.transform import ScaleByRStdDevState
from optax._src.transform import ScaleByScheduleState
from optax._src.transform import ScaleByTrustRatioState
from optax._src.transform import ScaleState
from optax._src.transform import trace
from optax._src.transform import TraceState
from optax._src.transform import TransformInitFn
from optax._src.transform import TransformUpdateFn
from optax._src.transform import Updates
from optax._src.transform import zero_nans
from optax._src.transform import ZeroNansState
from optax._src.update import apply_updates
from optax._src.update import incremental_update
from optax._src.update import periodic_update
from optax._src.utils import multi_normal
from optax._src.wrappers import apply_if_finite
from optax._src.wrappers import ApplyIfFiniteState
from optax._src.wrappers import flatten
from optax._src.wrappers import lookahead
from optax._src.wrappers import LookaheadParams
from optax._src.wrappers import LookaheadState
from optax._src.wrappers import masked
from optax._src.wrappers import MaskedState
from optax._src.wrappers import maybe_update
from optax._src.wrappers import MaybeUpdateState
from optax._src.wrappers import MultiSteps
from optax._src.wrappers import MultiStepsState

__version__ = "0.0.6"

__all__ = (
    "adabelief",
    "adagrad",
    "adam",
    "adamw",
    "adaptive_grad_clip",
    "add_decayed_weights",
    "add_noise",
    "AddDecayedWeightsState",
    "additive_weight_decay",
    "AdditiveWeightDecayState",
    "AddNoiseState",
    "apply_if_finite",
    "apply_every",
    "apply_updates",
    "ApplyEvery",
    "ApplyIfFiniteState",
    "centralize",
    "chain",
    "clip",
    "clip_by_global_norm",
    "ClipByGlobalNormState",
    "ClipState",
    "constant_schedule",
    "control_delta_method",
    "control_variates_jacobians",
    "cosine_decay_schedule",
    "cosine_distance",
    "cosine_onecycle_schedule",
    "dpsgd",
    "differentially_private_aggregate",
    "DifferentiallyPrivateAggregateState",
    "exponential_decay",
    "fisher_diag",
    "flatten",
    "fromage",
    "global_norm",
    "GradientTransformation",
    "hessian_diag",
    "huber_loss",
    "hvp",
    "identity",
    "incremental_update",
    "inject_hyperparams",
    "InjectHyperparamsState",
    "lamb",
    "lookahead",
    "LookaheadParams",
    "LookaheadState",
    "l2_loss",
    "linear_onecycle_schedule",
    "matrix_inverse_pth_root",
    "masked",
    "MaskedState",
    "measure_valued_jacobians",
    "moving_avg_baseline",
    "multi_normal",
    "noisy_sgd",
    "OptState",
    "Params",
    "pathwise_jacobians",
    "periodic_update",
    "piecewise_constant_schedule",
    "piecewise_interpolate_schedule",
    "polynomial_schedule",
    "power_iteration",
    "radam",
    "rmsprop",
    "scale",
    "scale_by_adam",
    "scale_by_belief",
    "scale_by_radam",
    "scale_by_rms",
    "scale_by_rss",
    "scale_by_schedule",
    "scale_by_stddev",
    "scale_by_trust_ratio",
    "scale_by_yogi",
    "ScaleByAdamState",
    "ScaleByFromageState",
    "ScaleByRmsState",
    "ScaleByRssState",
    "ScaleByRStdDevState",
    "ScaleByScheduleState",
    "ScaleByTrustRatioState",
    "ScaleState",
    "Schedule",
    "score_function_jacobians",
    "sgd",
    "sigmoid_binary_cross_entropy",
    "smooth_labels",
    "softmax_cross_entropy",
    "trace",
    "TraceState",
    "TransformInitFn",
    "TransformUpdateFn",
    "Updates",
    "yogi",
)

#  _________________________________________
# / Please don't use symbols in `_src` they \
# \ are not part of the Optax public API.   /
#  -----------------------------------------
#         \   ^__^
#          \  (oo)\_______
#             (__)\       )\/\
#                 ||----w |
#                 ||     ||
#



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Configuration file for the Sphinx documentation builder."""

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

# pylint: disable=g-bad-import-order
# pylint: disable=g-import-not-at-top
import inspect
import os
import sys
import typing


def _add_annotations_import(path):
  """Appends a future annotations import to the file at the given path."""
  with open(path) as f:
    contents = f.read()
  if contents.startswith('from __future__ import annotations'):
    # If we run sphinx multiple times then we will append the future import
    # multiple times too.
    return

  assert contents.startswith('#'), (path, contents.split('\n')[0])
  with open(path, 'w') as f:
    # NOTE: This is subtle and not unit tested, we're prefixing the first line
    # in each Python file with this future import. It is important to prefix
    # not insert a newline such that source code locations are accurate (we link
    # to GitHub). The assertion above ensures that the first line in the file is
    # a comment so it is safe to prefix it.
    f.write('from __future__ import annotations  ')
    f.write(contents)


def _recursive_add_annotations_import():
  for path, _, files in os.walk('../optax/'):
    for file in files:
      if file.endswith('.py'):
        _add_annotations_import(os.path.abspath(os.path.join(path, file)))

if 'READTHEDOCS' in os.environ:
  _recursive_add_annotations_import()

typing.get_type_hints = lambda obj, *unused: obj.__annotations__
sys.path.insert(0, os.path.abspath('../'))
sys.path.append(os.path.abspath('ext'))

import optax
import sphinxcontrib.katex as katex

# -- Project information -----------------------------------------------------

project = 'Optax'
copyright = '2021, DeepMind'  # pylint: disable=redefined-builtin
author = 'Optax Contributors'

# -- General configuration ---------------------------------------------------

master_doc = 'index'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.intersphinx',
    'sphinx.ext.linkcode',
    'sphinx.ext.napoleon',
    'sphinxcontrib.bibtex',
    'sphinxcontrib.katex',
    'sphinx_autodoc_typehints',
    'coverage_check',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Options for autodoc -----------------------------------------------------

autodoc_default_options = {
    'member-order': 'bysource',
    'special-members': True,
    'exclude-members': '__repr__, __str__, __weakref__',
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
# html_favicon = '_static/favicon.ico'

# -- Options for katex ------------------------------------------------------

# See: https://sphinxcontrib-katex.readthedocs.io/en/0.4.1/macros.html
latex_macros = r"""
    \def \d              #1{\operatorname{#1}}
"""

# Translate LaTeX macros to KaTeX and add to options for HTML builder
katex_macros = katex.latex_defs_to_katex_macros(latex_macros)
katex_options = 'macros: {' + katex_macros + '}'

# Add LaTeX macros for LATEX builder
latex_elements = {'preamble': latex_macros}

# -- Source code links -------------------------------------------------------


def linkcode_resolve(domain, info):
  """Resolve a GitHub URL corresponding to Python object."""
  if domain != 'py':
    return None

  try:
    mod = sys.modules[info['module']]
  except ImportError:
    return None

  obj = mod
  try:
    for attr in info['fullname'].split('.'):
      obj = getattr(obj, attr)
  except AttributeError:
    return None
  else:
    obj = inspect.unwrap(obj)

  try:
    filename = inspect.getsourcefile(obj)
  except TypeError:
    return None

  try:
    source, lineno = inspect.getsourcelines(obj)
  except OSError:
    return None

  # TODO(slebedev): support tags after we release an initial version.
  return 'https://github.com/deepmind/optax/tree/master/optax/%s#L%d#L%d' % (
      os.path.relpath(filename, start=os.path.dirname(
          optax.__file__)), lineno, lineno + len(source) - 1)


# -- Intersphinx configuration -----------------------------------------------

intersphinx_mapping = {
    'jax': ('https://jax.readthedocs.io/en/latest/', None),
}




# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX Schedules.

Schedules may be used to anneal the value of a hyper-parameter over time; for
instance, they may be used to anneal the learning rate used to update an agent's
parameters or the exploration factor used to select actions.
"""

import functools
import inspect
from typing import Callable, Dict, Union, Optional, Iterable

from absl import logging
import chex
import jax
import jax.numpy as jnp

from optax._src import transform
from optax._src import utils

Schedule = Callable[[chex.Numeric], chex.Numeric]


def constant_schedule(value: Union[float, int]) -> Schedule:
  """Constructs a constant schedule.

  Args:
    value: value to be held constant throughout.

  Returns:
    schedule: A function that maps step counts to values.
  """
  return lambda count: value


def polynomial_schedule(
    init_value: chex.Scalar,
    end_value: chex.Scalar,
    power: chex.Scalar,
    transition_steps: int,
    transition_begin: int = 0) -> Schedule:
  """Constructs a schedule with polynomial transition from init to end value.

  Args:
    init_value: initial value for the scalar to be annealed.
    end_value: end value of the scalar to be annealed.
    power: the power of the polynomial used to transition from init to end.
    transition_steps: number of steps over which annealing takes place,
      the scalar starts changing at `transition_begin` steps and completes
      the transition by `transition_begin + transition_steps` steps.
      If `transition_steps <= 0`, then the entire annealing process is disabled
      and the value is held fixed at `init_value`.
    transition_begin: must be positive. After how many steps to start annealing
      (before this many steps the scalar value is held fixed at `init_value`).

  Returns:
    schedule: A function that maps step counts to values.
  """
  if transition_steps <= 0:
    logging.info(
        'A polynomial schedule was set with a non-positive `transition_steps` '
        'value; this results in a constant schedule with value `init_value`.')
    return lambda count: init_value

  if transition_begin < 0:
    logging.info(
        'An exponential schedule was set with a negative `transition_begin` '
        'value; this will result in `transition_begin` falling back to `0`.')
    transition_begin = 0

  def schedule(count):
    count = jnp.clip(count - transition_begin, 0, transition_steps)
    frac = 1 - count / transition_steps
    return (init_value - end_value) * (frac**power) + end_value
  return schedule


def piecewise_constant_schedule(
    init_value: float,
    boundaries_and_scales: Optional[Dict[int, float]] = None) -> Schedule:
  """Returns a function which implements a piecewise constant schedule.

  Args:
    init_value: An initial value `init_v`.
    boundaries_and_scales: A map from boundaries `b_i` to non-negative scaling
      factors `f_i`. For any step count `s`, the schedule returns `init_v`
      scaled by the product of all factors `f_i` such that `b_i` < `s`.

  Returns:
    schedule: A function that maps step counts to values.
  """
  if boundaries_and_scales is not None:
    all_positive = all(scale >= 0. for scale in boundaries_and_scales.values())
    if not all_positive:
      raise ValueError(
          '`piecewise_constant_schedule` expects non-negative scale factors')

  def schedule(count):
    v = init_value
    if boundaries_and_scales is not None:
      for threshold, scale in sorted(boundaries_and_scales.items()):
        indicator = jnp.maximum(0., jnp.sign(threshold - count))
        v = v * indicator + (1 - indicator) * scale * v
    return v

  return schedule


def exponential_decay(
    init_value: float,
    transition_steps: int,
    decay_rate: float,
    transition_begin: int = 0,
    staircase: bool = False,
    end_value: Optional[float] = None) -> Schedule:
  """Constructs a schedule with either continuous or discrete exponential decay.

  This function applies an exponential decay function to a provided initial
  value. The function returns the decayed value as follows:

  ```
  decayed_value = init_value * decay_rate ^ (count / transition_steps)
  ```

  If the argument `staircase` is `True`, then `count / transition_steps` is
  an integer division and the decayed value follows a staircase function.

  Args:
    init_value: the initial learning rate.
    transition_steps: must be positive. See the decay computation above.
    decay_rate: must not be zero. The decay rate.
    transition_begin: must be positive. After how many steps to start annealing
      (before this many steps the scalar value is held fixed at `init_value`).
    staircase: if `True`, decay the values at discrete intervals.
    end_value: the value at which the exponential decay stops. When
      `decay_rate` < 1, `end_value` is treated as a lower bound, otherwise as
      an upper bound. Has no effect when `decay_rate` = 0.

  Returns:
    schedule: A function that maps step counts to values.
  """

  if transition_steps <= 0:
    logging.info(
        'An exponential schedule was set with a non-positive `transition_steps`'
        ' value; this will result in a constant schedule with value '
        '`init_value`.')
    return lambda count: init_value

  if decay_rate == 0:
    logging.info(
        'An exponential schedule was set with a zero `decay_rate` value; '
        'this will result in a constant schedule with value `init_value`.')
    return lambda count: init_value

  if transition_begin < 0:
    logging.info(
        'An exponential schedule was set with a negative `transition_begin` '
        'value; this will result in `transition_begin` falling back to `0`.')
    transition_begin = 0

  if end_value is not None:
    clip_fn = jnp.maximum if decay_rate < 1.0 else jnp.minimum

  def schedule(count):
    count -= transition_begin
    p = count / transition_steps
    if staircase:
      p = jnp.floor(p)
    decayed_value = jnp.where(
        count <= 0, init_value, init_value * jnp.power(decay_rate, p))
    if end_value is not None:
      decayed_value = clip_fn(decayed_value, end_value)
    return decayed_value

  return schedule


def cosine_decay_schedule(
    init_value: float,
    decay_steps: int,
    alpha: float = 0.0) -> Schedule:
  """Returns a function which implements cosine learning rate decay.

  For more details see: https://arxiv.org/abs/1608.03983

  Args:
    init_value: An initial value `init_v`.
    decay_steps: Positive integer - the number of steps for which to apply
      the decay for.
    alpha: Float. The minimum value of the multiplier used to adjust the
      learning rate.

  Returns:
    schedule: A function that maps step counts to values.
  """
  if not decay_steps > 0:
    raise ValueError('The cosine_decay_schedule requires positive decay_steps!')

  def schedule(count):
    count = jnp.minimum(count, decay_steps)
    cosine_decay = 0.5 * (1 + jnp.cos(jnp.pi * count / decay_steps))
    decayed = (1 - alpha) * cosine_decay + alpha
    return init_value * decayed

  return schedule


def _linear_interpolate(start: float, end: float, pct: float):
  return (end-start) * pct + start


def _cosine_interpolate(start: float, end: float, pct: float):
  return end + (start-end) / 2.0 * (jnp.cos(jnp.pi * pct) + 1)


def piecewise_interpolate_schedule(
    interpolate_type: str,
    init_value: float,
    boundaries_and_scales: Optional[Dict[int, float]] = None) -> Schedule:
  """Returns a function which implements a piecewise interpolated schedule.

  Args:
    interpolate_type: 'linear' or 'cosine', specifying the interpolation
      strategy.
    init_value: An initial value `init_v`.
    boundaries_and_scales: A map from boundaries `b_i` to non-negative scaling
      factors `f_i`. At boundary step `b_i`, the schedule returns `init_v`
      scaled by the product of all factors `f_j` such that `b_j` < `b_i`. The
      values in between each boundary will be interpolated as per `type`.

  Returns:
    schedule: A function that maps step counts to values.
  """
  if interpolate_type == 'linear':
    interpolate_fn = _linear_interpolate
  elif interpolate_type == 'cosine':
    interpolate_fn = _cosine_interpolate
  else:
    raise ValueError('`interpolate_type` must be either \'cos\' or \'linear\'')

  if boundaries_and_scales:
    boundaries, scales = zip(*sorted(boundaries_and_scales.items()))
    if not all(scale >= 0. for scale in scales):
      raise ValueError(
          '`piecewise_interpolate_schedule` expects non-negative scale factors')
  else:
    boundaries, scales = (), ()

  bounds = jnp.stack((0,) + boundaries)
  values = jnp.cumprod(jnp.stack((init_value,) + scales))
  interval_sizes = (bounds[1:] - bounds[:-1])

  def schedule(count):
    indicator = (bounds[:-1] <= count) & (count < bounds[1:])
    pct = (count - bounds[:-1]) / interval_sizes
    interp_vals = interpolate_fn(values[:-1], values[1:], pct)
    return indicator.dot(interp_vals) + (bounds[-1] <= count) * values[-1]

  return schedule


def linear_onecycle_schedule(
    transition_steps: int,
    peak_value: float,
    pct_start: float = 0.3,
    pct_final: float = 0.85,
    div_factor: float = 25.0,
    final_div_factor: float = 1e4) -> Schedule:
  """Returns a function which implements the onecycle learning rate schedule.

  This function uses a linear annealing strategy.
  For more details see: https://arxiv.org/abs/1708.07120

  Args:
    transition_steps: Number of steps over which annealing takes place.
    peak_value: Maximum value attained by schedule at pct_start percent
      of the cycle (in number of steps).
    pct_start: The percentage of the cycle (in number of steps) spent
      increasing the learning rate.
    pct_final: The percentage of the cycle (in number of steps) spent
      increasing to peak_value then decreasing back to init_value.
    div_factor: Determines the initial value via init_value =
      peak_value / div_factor
    final_div_factor: Determines the final value via final_value =
      init_value / final_div_factor

  Returns:
    schedule: A function that maps step counts to values.
  """
  if transition_steps <= 0:
    raise ValueError(
        'A linear onecycle schedule was set with a non-positive '
        '`transition_steps`')

  return piecewise_interpolate_schedule(
      'linear',
      peak_value / div_factor,
      {int(pct_start * transition_steps): div_factor,
       int(pct_final * transition_steps): 1. / div_factor,
       transition_steps: 1. / final_div_factor})


def cosine_onecycle_schedule(
    transition_steps: int,
    peak_value: float,
    pct_start: float = 0.3,
    div_factor: float = 25.0,
    final_div_factor: float = 1e4) -> Schedule:
  """Returns a function which implements the onecycle learning rate schedule.

  This function uses a cosine annealing strategy.
  For more details see: https://arxiv.org/abs/1708.07120

  Args:
    transition_steps: Number of steps over which annealing takes place.
    peak_value: Maximum value attained by schedule at pct_start percent
      of the cycle (in number of steps).
    pct_start: The percentage of the cycle (in number of steps) spent
      increasing the learning rate.
    div_factor: Determines the initial value via init_value =
      peak_value / div_factor
    final_div_factor: Determines the final value via final_value =
      init_value / final_div_factor

  Returns:
    schedule: A function that maps step counts to values.
  """
  if transition_steps <= 0:
    raise ValueError(
        'A linear onecycle schedule was set with a non-positive '
        '`transition_steps`')

  return piecewise_interpolate_schedule(
      'cosine',
      peak_value / div_factor,
      {int(pct_start * transition_steps): div_factor,
       int(transition_steps): 1. / (div_factor * final_div_factor)})


def _convert_floats(x, dtype):
  """Convert float-like inputs to dtype, rest pass through."""
  if jax.dtypes.scalar_type_of(x) == float:
    return jnp.asarray(x, dtype=dtype)
  return x


class InjectHyperparamsState(transform.OptState):
  """Maintains inner transform state, hyperparameters, and step count."""
  count: jnp.ndarray  # shape=(), dtype=jnp.int32
  hyperparams: Dict[str, chex.Numeric]
  inner_state: transform.OptState


def inject_hyperparams(
    inner_factory: Callable[..., transform.GradientTransformation],
    static_args: Union[str, Iterable[str]] = ()
) -> Callable[..., transform.GradientTransformation]:
  """Wrapper that injects hyperparameters into the inner GradientTransformation.

  This wrapper allows you to pass schedules (i.e. a function that returns a
  numeric value given a step count) instead of constants for
  hyperparameters. You may only schedule numeric hyperparameters (i.e. boolean
  flags cannot be scheduled).

  For example, to use `scale_by_adam` with a piecewise linear
  schedule for beta_1 and constant for beta_2:

  ```
  scheduled_adam = optax.inject_hyperparams(optax.scale_by_adam)(
      b1=optax.piecewise_linear_schedule(...),
      b2=0.99)
  ```

  You may manually change numeric hyperparameters that were not scheduled
  through the `hyperparams` dict in the InjectHyperparamState:

  ```
  state = scheduled_adam.init(params)
  updates, state = scheduled_adam.update(grads, state)
  state.hyperparams['b2'] = 0.95
  updates, state = scheduled_adam.update(updates, state)  # uses b2 = 0.95
  ```

  Manually overriding scheduled hyperparameters will have no effect (e.g.
  in the code sample above, you cannot manually adjust `b1`).

  Args:
    inner_factory: a function that returns the inner
      `optax.GradientTransformation` given the hyperparameters.
    static_args: a string or iterable of strings specifying which
      callable parameters are not schedules. inject_hyperparams treats all
      callables as schedules by default, so if a hyperparameter is a
      non-schedule callable, you must specify that using this argument.

  Returns:
    A callable that returns a `optax.GradientTransformation`. This callable
    accepts the same arguments as `inner_factor`, except you may provide
    schedules in place of the constant arguments.
  """
  static_args = ({static_args} if isinstance(static_args, str) else
                 set(static_args))
  inner_signature = inspect.signature(inner_factory)

  if not static_args.issubset(inner_signature.parameters):
    raise ValueError(
        '`static_args` must specify a subset of `inner_factory`\'s parameters. '
        f'Given `static_args`: {static_args}. `inner_factory` parameters: '
        f'{set(inner_signature.parameters.keys())}')

  @functools.wraps(inner_factory)
  def wrapped_transform(*args, **kwargs) -> transform.GradientTransformation:
    bound_arguments = inner_signature.bind(*args, **kwargs)
    bound_arguments.apply_defaults()

    sched_hps, numeric_hps, other_hps = {}, {}, {}
    for name, value in bound_arguments.arguments.items():
      if name in static_args or isinstance(value, bool):
        other_hps[name] = value
      elif callable(value):
        sched_hps[name] = value
      elif isinstance(value, (int, float, chex.Array)):
        numeric_hps[name] = value
      else:
        other_hps[name] = value

    def schedule_fn(count, dtype):
      return {k: _convert_floats(f(count), dtype) for k, f in sched_hps.items()}

    def init_fn(params):
      count = jnp.zeros([], jnp.int32)
      dtype = getattr(next(iter(jax.tree_leaves(params)), None), 'dtype', None)
      hparams = {
          k: jnp.asarray(_convert_floats(v, dtype))
          for k, v in numeric_hps.items()}
      hparams.update(schedule_fn(count, dtype))
      return InjectHyperparamsState(  # pylint:disable=too-many-function-args
          count, hparams, inner_factory(**other_hps, **hparams).init(params))

    def update_fn(updates, state, params=None):
      count_inc = utils.safe_int32_increment(state.count)
      dtype = getattr(next(iter(jax.tree_leaves(updates)), None), 'dtype', None)
      hparams = {k: _convert_floats(v, dtype)
                 for k, v in state.hyperparams.items()}
      hparams.update(schedule_fn(count_inc, dtype))
      updates, inner_state = inner_factory(**other_hps, **hparams).update(
          updates, state.inner_state, params)

      # pylint:disable=too-many-function-args
      return updates, InjectHyperparamsState(count_inc, hparams, inner_state)
      # pylint:enable=too-many-function-args

    return transform.GradientTransformation(init_fn, update_fn)

  return wrapped_transform



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `transform.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
import jax
import jax.numpy as jnp

from optax._src import alias
from optax._src import combine
from optax._src import transform
from optax._src import update

STEPS = 50
LR = 1e-2


class TransformTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.init_params = (jnp.array([1., 2.]), jnp.array([3., 4.]))
    self.per_step_updates = (jnp.array([500., 5.]), jnp.array([300., 3.]))

  @chex.all_variants
  @parameterized.named_parameters([
      ('adam', transform.scale_by_adam),
      ('rmsprop', transform.scale_by_rms),
      ('stddev', transform.scale_by_stddev),
      ('trust_ratio', transform.scale_by_trust_ratio),
  ])
  def test_scalers(self, scaler_constr):
    params = self.init_params

    scaler = scaler_constr()
    init_fn = self.variant(scaler.init)
    transform_fn = self.variant(scaler.update)

    state = init_fn(params)
    chex.assert_tree_all_finite(state)

    updates, state = transform_fn(self.per_step_updates, state, params)
    chex.assert_tree_all_finite((params, updates, state))
    jax.tree_multimap(lambda *args: chex.assert_equal_shape(args), params,
                      updates)

  @chex.all_variants()
  def test_apply_every(self):
    # The frequency of the application of sgd
    k = 4
    zero_update = (jnp.array([0., 0.]), jnp.array([0., 0.]))

    # optax sgd
    optax_sgd_params = self.init_params
    sgd = alias.sgd(LR, 0.0)
    state_sgd = sgd.init(optax_sgd_params)

    # optax sgd plus apply every
    optax_sgd_apply_every_params = self.init_params
    sgd_apply_every = combine.chain(
        transform.apply_every(k=k),
        transform.trace(decay=0, nesterov=False),
        transform.scale(-LR))
    state_sgd_apply_every = sgd_apply_every.init(optax_sgd_apply_every_params)
    transform_fn = self.variant(sgd_apply_every.update)

    for i in range(STEPS):
      # Apply a step of sgd
      updates_sgd, state_sgd = sgd.update(self.per_step_updates, state_sgd)
      optax_sgd_params = update.apply_updates(optax_sgd_params, updates_sgd)

      # Apply a step of sgd_apply_every
      updates_sgd_apply_every, state_sgd_apply_every = transform_fn(
          self.per_step_updates, state_sgd_apply_every)
      optax_sgd_apply_every_params = update.apply_updates(
          optax_sgd_apply_every_params, updates_sgd_apply_every)

      # Every k steps, check equivalence.
      if i % k == k-1:
        chex.assert_tree_all_close(
            optax_sgd_apply_every_params, optax_sgd_params,
            atol=1e-6, rtol=1e-5)
      # Otherwise, check update is zero.
      else:
        chex.assert_tree_all_close(
            updates_sgd_apply_every, zero_update, atol=0.0, rtol=0.0)

  def test_scale(self):
    updates = self.per_step_updates
    for i in range(1, STEPS + 1):
      factor = 0.1 ** i
      rescaler = transform.scale(factor)
      # Apply rescaling.
      scaled_updates, _ = rescaler.update(updates, None)
      # Manually scale updates.
      def rescale(t):
        return t * factor  # pylint:disable=cell-var-from-loop
      manual_updates = jax.tree_map(rescale, updates)
      # Check the rescaled updates match.
      chex.assert_tree_all_close(scaled_updates, manual_updates)

  def test_clip(self):
    updates = self.per_step_updates
    # For a sufficiently high delta the update should not be changed.
    clipper = transform.clip(1e6)
    clipped_updates, _ = clipper.update(updates, None)
    chex.assert_tree_all_close(clipped_updates, clipped_updates)
    # Clipping at delta=1 should make all updates exactly 1.
    clipper = transform.clip(1.)
    clipped_updates, _ = clipper.update(updates, None)
    chex.assert_tree_all_close(
        clipped_updates, jax.tree_map(jnp.ones_like, updates))

  def test_clip_by_global_norm(self):
    updates = self.per_step_updates
    for i in range(1, STEPS + 1):
      clipper = transform.clip_by_global_norm(1. / i)
      updates, _ = clipper.update(updates, None)
      # Check that the clipper actually works and global norm is <= max_norm
      self.assertAlmostEqual(transform.global_norm(updates), 1. / i, places=6)
      updates_step, _ = clipper.update(self.per_step_updates, None)
      # Check that continuously clipping won't cause numerical issues.
      chex.assert_tree_all_close(updates, updates_step, atol=1e-7, rtol=1e-7)

  @parameterized.named_parameters([
      ('1d', [1.0, 2.0], [1.0, 2.0]),
      ('2d', [[1.0, 2.0], [3.0, 4.0]], [[-0.5, 0.5], [-0.5, 0.5]]),
      ('3d', [[[1., 2.], [3., 4.]],
              [[5., 6.], [7., 8.]]], [[[-1.5, -0.5], [0.5, 1.5]],
                                      [[-1.5, -0.5], [0.5, 1.5]]]),
  ])
  def test_centralize(self, inputs, outputs):
    inputs = jnp.asarray(inputs)
    outputs = jnp.asarray(outputs)
    centralizer = transform.centralize()
    centralized_inputs, _ = centralizer.update(inputs, None)
    chex.assert_tree_all_close(centralized_inputs, outputs)

  def test_keep_params_nonnegative(self):
    grads = (jnp.array([500., -500., 0.]),
             jnp.array([500., -500., 0.]),
             jnp.array([500., -500., 0.]))

    params = (jnp.array([-1., -1., -1.]),
              jnp.array([1., 1., 1.]),
              jnp.array([0., 0., 0.]))

    # vanilla sgd
    opt = combine.chain(
        transform.trace(decay=0, nesterov=False), transform.scale(-LR))
    opt_state = opt.init(params)

    updates, _ = opt.update(grads, opt_state, params)
    new_params = update.apply_updates(params, updates)

    chex.assert_tree_all_close(new_params, (jnp.array([-6., 4., -1.]),
                                            jnp.array([-4., 6., 1.]),
                                            jnp.array([-5., 5., 0.])))

    # sgd with keeping parameters non-negative
    opt = combine.chain(
        transform.trace(decay=0, nesterov=False), transform.scale(-LR),
        transform.keep_params_nonnegative())
    opt_state = opt.init(params)

    updates, _ = opt.update(grads, opt_state, params)
    new_params = update.apply_updates(params, updates)

    chex.assert_tree_all_close(new_params, (jnp.array([0., 4., 0.]),
                                            jnp.array([0., 6., 1.]),
                                            jnp.array([0., 5., 0.])))

  @chex.all_variants
  def test_zero_nans(self):
    params = (jnp.zeros([3]), jnp.zeros([3]), jnp.zeros([3]))

    opt = transform.zero_nans()
    opt_state = self.variant(opt.init)(params)
    update_fn = self.variant(opt.update)

    equality_comp = lambda a, b: bool(jnp.all(jnp.equal(a, b)))
    chex.assert_tree_all_equal_comparator(equality_comp, opt_state,
                                          (jnp.array(False),) * 3)

    # Check an upate with nans
    grads_with_nans = (jnp.ones([3]),
                       jnp.array([1., float('nan'), float('nan')]),
                       jnp.array([float('nan'), 1., 1.]))
    updates, opt_state = update_fn(grads_with_nans, opt_state)
    chex.assert_tree_all_equal_comparator(
        equality_comp, opt_state,
        (jnp.array(False), jnp.array(True), jnp.array(True)))
    chex.assert_tree_all_equal_comparator(
        equality_comp, updates,
        (jnp.ones([3]), jnp.array([1., 0., 0.]), jnp.array([0., 1., 1.])))

    # Check an upate with nans and infs
    grads_with_nans_infs = (jnp.ones([3]),
                            jnp.array([1., float('nan'), float('nan')]),
                            jnp.array([float('inf'), 1., 1.]))
    updates, opt_state = update_fn(grads_with_nans_infs, opt_state)
    chex.assert_tree_all_equal_comparator(
        equality_comp, opt_state,
        (jnp.array(False), jnp.array(True), jnp.array(False)))
    chex.assert_tree_all_equal_comparator(
        equality_comp, updates,
        (jnp.ones([3]), jnp.array([1., 0., 0.]),
         jnp.array([float('inf'), 1., 1.])))

    # Check an upate with only good values
    grads = (jnp.ones([3]), jnp.ones([3]), jnp.ones([3]))
    updates, opt_state = update_fn(grads, opt_state)
    chex.assert_tree_all_equal_comparator(
        equality_comp, opt_state,
        (jnp.array(False), jnp.array(False), jnp.array(False)))
    chex.assert_tree_all_equal_comparator(equality_comp, updates, grads)


if __name__ == '__main__':
  absltest.main()



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
r"""Stochastic Monte Carlo gradient estimators.

Utility functions to approximate gradients of the form using Monte Carlo
estimation:
  \nabla_{\theta} E_{p(x; \theta)} f(x)

Here f is assumed to have no dependence on the parameters theta - if f has
dependence on theta, the functions below need to be called with `stop_grad(f)`
and the chain rule needs to be applied outside these functions in order
to obtain unbiased gradient.

For more details, see:
S. Mohamed, M. Rosca, M. Figurnov, A Mnih.
  Monte Carlo Gradient Estimation in Machine Learning. JMLR, 2020.
"""

import math
from typing import Any, Callable, Sequence

import chex
import jax
import jax.numpy as jnp
import numpy as np
from optax._src import transform
from optax._src import utils


def score_function_jacobians(
    function: Callable[[chex.Array], float],
    params: transform.Params,
    dist_builder: Callable[..., Any],
    rng: chex.PRNGKey,
    num_samples: int) -> Sequence[chex.Array]:
  r"""Score function gradient estimation.

  Approximates:
     \nabla_{\theta} E_{p(x; \theta)} f(x)
  With:
    E_{p(x; \theta)} f(x) \nabla_{\theta} \log p(x; \theta)

  Requires: p to be differentiable wrt to theta. Applicable to both continuous
    and discrete random variables. No requirements on f.

  Args:
    function: Function f(x) for which to estimate grads_{params} E_dist f(x).
      The function takes in one argument (a sample from the distribution) and
      returns a floating point value.
    params: A tuple of jnp arrays.
      The parameters for which to construct the distribution.
    dist_builder: a constructor which builds a distribution given the input
      parameters specified by params. `dist_builder(params)` should return a
      valid distribution.
    rng: a PRNGKey key.
    num_samples: Int, the number of samples used to compute the grads.

  Returns:
    A tuple of size `params`, each element is `num_samples x param.shape`
      jacobian vector containing the estimates of the gradients obtained for
      each sample.
    The mean of this vector is the gradient wrt to parameters that can be used
      for learning. The entire jacobian vector can be used to assess estimator
      variance.
  """
  def surrogate(params):
    dist = dist_builder(*params)
    one_sample_surrogate_fn = lambda x: function(x) * dist.log_prob(x)
    samples = jax.lax.stop_gradient(dist.sample((num_samples,), seed=rng))
    # We vmap the function application over samples - this ensures that the
    # function we use does not have to be vectorized itself.
    return jax.vmap(one_sample_surrogate_fn)(samples)

  return jax.jacfwd(surrogate)(params)


def pathwise_jacobians(
    function: Callable[[chex.Array], float],
    params: transform.Params,
    dist_builder: Callable[..., Any],
    rng: chex.PRNGKey,
    num_samples: int) -> Sequence[chex.Array]:
  r"""Pathwise gradient estimation.

  Approximates:
     \nabla_{\theta} E_{p(x; \theta)} f(x)
  With:
    E_{p(\epsilon)} \nabla_{\theta} f(g(\epsilon, \theta))
      where x = g(\epsilon, \theta). g depends on the distribution p.

  Requires: p to be reparametrizable and the reparametrization to be implemented
    in tensorflow_probability. Applicable to continuous random variables.
    f needs to be differentiable.

  Args:
    function: Function f(x) for which to estimate grads_{params} E_dist f(x).
      The function takes in one argument (a sample from the distribution) and
      returns a floating point value.
    params: A tuple of jnp arrays.
      The parameters for which to construct the distribution.
    dist_builder: a constructor which builds a distribution given the input
      parameters specified by params. `dist_builder(params)` should return a
      valid distribution.
    rng: a PRNGKey key.
    num_samples: Int, the number of samples used to compute the grads.

  Returns:
    A tuple of size `params`, each element is `num_samples x param.shape`
      jacobian vector containing the estimates of the gradients obtained for
      each sample.
    The mean of this vector is the gradient wrt to parameters that can be used
      for learning. The entire jacobian vector can be used to assess estimator
      variance.
  """
  def surrogate(params):
    # We vmap the function application over samples - this ensures that the
    # function we use does not have to be vectorized itself.
    dist = dist_builder(*params)
    return jax.vmap(function)(dist.sample((num_samples,), seed=rng))

  return jax.jacfwd(surrogate)(params)


def measure_valued_jacobians(
    function: Callable[[chex.Array], float],
    params: transform.Params,
    dist_builder: Callable[..., Any],
    rng: chex.PRNGKey,
    num_samples: int,
    coupling: bool = True) -> Sequence[chex.Array]:
  r"""Measure valued gradient estimation.

  Approximates:
     \nabla_{\theta} E_{p(x; \theta)} f(x)
  With:
    1./ c (E_{p1(x; \theta)} f(x) - E_{p2(x; \theta)} f(x)) where p1 and p2 are
    measures which depend on p.

  Currently only supports computing gradients of expectations of Gaussian RVs.

  Args:
    function: Function f(x) for which to estimate grads_{params} E_dist f(x).
      The function takes in one argument (a sample from the distribution) and
      returns a floating point value.
    params: A tuple of jnp arrays.
      The parameters for which to construct the distribution.
    dist_builder: a constructor which builds a distribution given the input
      parameters specified by params. `dist_builder(params)` should return a
      valid distribution.
    rng: a PRNGKey key.
    num_samples: Int, the number of samples used to compute the grads.
    coupling: A boolean. Whether or not to use coupling for the positive and
      negative samples. Recommended: True, as this reduces variance.

  Returns:
    A tuple of size `params`, each element is `num_samples x param.shape`
      jacobian vector containing the estimates of the gradients obtained for
      each sample.
    The mean of this vector is the gradient wrt to parameters that can be used
      for learning. The entire jacobian vector can be used to assess estimator
      variance.
  """
  if dist_builder is not utils.multi_normal:
    raise ValueError(
        'Unsupported distribution builder for measure_valued_jacobians!')
  dist = dist_builder(*params)
  # Need to apply chain rule for log scale grad (instead of scale grad).
  return [
      measure_valued_estimation_mean(
          function, dist, rng, num_samples, coupling=coupling),
      jnp.exp(dist.log_scale) * measure_valued_estimation_std(
          function, dist, rng, num_samples, coupling=coupling)]


def measure_valued_estimation_mean(
    function: Callable[[chex.Array], float],
    dist: Any,
    rng: chex.PRNGKey,
    num_samples: int,
    coupling: bool = True) -> chex.Array:
  """Measure valued grads of a Gaussian expectation of `function` wrt the mean.

  Args:
    function: Function f(x) for which to estimate grads_{mean} E_dist f(x).
      The function takes in one argument (a sample from the distribution) and
      returns a floating point value.
    dist: a distribution on which we can call `sample`.
    rng: a PRNGKey key.
    num_samples: Int, the number of samples used to compute the grads.
    coupling: A boolean. Whether or not to use coupling for the positive and
      negative samples. Recommended: True, as this reduces variance.

  Returns:
    A `num_samples x D` vector containing the estimates of the gradients
    obtained for each sample. The mean of this vector can be used to update
    the mean parameter. The entire vector can be used to assess estimator
    variance.
  """
  mean, log_std = dist.params
  std = jnp.exp(log_std)

  dist_samples = dist.sample((num_samples,), seed=rng)

  pos_rng, neg_rng = jax.random.split(rng)
  pos_sample = jax.random.weibull_min(
      pos_rng, scale=math.sqrt(2.), concentration=2., shape=dist_samples.shape)

  if coupling:
    neg_sample = pos_sample
  else:
    neg_sample = jax.random.weibull_min(
        neg_rng,
        scale=math.sqrt(2.),
        concentration=2.,
        shape=dist_samples.shape)

  # N x D
  positive_diag = mean + std * pos_sample
  # N x D
  negative_diag = mean - std * neg_sample

  # NOTE: you can sample base samples here if you use the same rng
  # Duplicate the D dimension - N x D x D.
  base_dist_samples = utils.tile_second_to_last_dim(dist_samples)
  positive = utils.set_diags(base_dist_samples, positive_diag)
  negative = utils.set_diags(base_dist_samples, negative_diag)

  c = np.sqrt(2 * np.pi) * std  # D
  # Apply function. We apply the function to each element of N x D x D.
  # We apply a function that takes a sample and returns one number, so the
  # output will be N x D (which is what we want, batch by dimension).
  # We apply a function in parallel to the batch.
  # Broadcast the division.
  vmaped_function = jax.vmap(jax.vmap(function, 1, 0))
  grads = (vmaped_function(positive) - vmaped_function(negative)) / c

  chex.assert_shape(grads, (num_samples,) + std.shape)
  return grads


def measure_valued_estimation_std(
    function: Callable[[chex.Array], float],
    dist: Any,
    rng: chex.PRNGKey,
    num_samples: int,
    coupling: bool = True) -> chex.Array:
  """Measure valued grads of a Gaussian expectation of `function` wrt the std.

  Args:
    function: Function f(x) for which to estimate grads_{std} E_dist f(x).
      The function takes in one argument (a sample from the distribution) and
      returns a floating point value.
    dist: a distribution on which we can call `sample`.
    rng: a PRNGKey key.
    num_samples: Int, the number of samples used to compute the grads.
    coupling: A boolean. Whether or not to use coupling for the positive and
      negative samples. Recommended: True, as this reduces variance.

  Returns:
    A `num_samples x D` vector containing the estimates of the gradients
    obtained for each sample. The mean of this vector can be used to update
    the scale parameter. The entire vector can be used to assess estimator
    variance.
  """
  mean, log_std = dist.params
  std = jnp.exp(log_std)

  dist_samples = dist.sample((num_samples,), seed=rng)

  pos_rng, neg_rng = jax.random.split(rng)

  # The only difference between mean and std gradients is what we sample.
  pos_sample = jax.random.double_sided_maxwell(
      pos_rng, loc=0.0, scale=1.0, shape=dist_samples.shape)
  if coupling:
    unif_rvs = jax.random.uniform(neg_rng, dist_samples.shape)
    neg_sample = unif_rvs * pos_sample
  else:
    neg_sample = jax.random.normal(neg_rng, dist_samples.shape)

  # Both need to be positive in the case of the scale.
  # N x D
  positive_diag = mean + std * pos_sample
  # N x D
  negative_diag = mean + std * neg_sample

  # NOTE: you can sample base samples here if you use the same rng
  # Duplicate the D dimension - N x D x D.
  base_dist_samples = utils.tile_second_to_last_dim(dist_samples)
  positive = utils.set_diags(base_dist_samples, positive_diag)
  negative = utils.set_diags(base_dist_samples, negative_diag)

  # Different C for the scale
  c = std  # D
  # Apply function. We apply the function to each element of N x D x D.
  # We apply a function that takes a sample and returns one number, so the
  # output will be N x D (which is what we want, batch by dimension).
  # We apply a function in parallel to the batch.
  # Broadcast the division.
  vmaped_function = jax.vmap(jax.vmap(function, 1, 0))
  grads = (vmaped_function(positive) - vmaped_function(negative)) / c

  chex.assert_shape(grads, (num_samples,) + std.shape)
  return grads




# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `update.py`."""

from absl.testing import absltest
import chex
import jax
import jax.numpy as jnp
from optax._src import update


class UpdateTest(chex.TestCase):

  @chex.all_variants()
  def test_apply_updates(self):
    params = ({'a': jnp.ones((3, 2))}, jnp.ones((1,)))
    grads = jax.tree_map(lambda t: 2 * t, params)
    exp_params = jax.tree_map(lambda t: 3 * t, params)
    new_params = self.variant(update.apply_updates)(params, grads)

    chex.assert_tree_all_close(
        exp_params, new_params, atol=1e-10, rtol=1e-5)

  @chex.all_variants()
  def test_incremental_update(self):
    params_1 = ({'a': jnp.ones((3, 2))}, jnp.ones((1,)))
    params_2 = jax.tree_map(lambda t: 2 * t, params_1)
    exp_params = jax.tree_map(lambda t: 1.5 * t, params_1)
    new_params = self.variant(
        update.incremental_update)(params_2, params_1, 0.5)

    chex.assert_tree_all_close(
        exp_params, new_params, atol=1e-10, rtol=1e-5)

  @chex.all_variants()
  def test_periodic_update(self):
    params_1 = ({'a': jnp.ones((3, 2))}, jnp.ones((1,)))
    params_2 = jax.tree_map(lambda t: 2 * t, params_1)

    update_period = 5
    update_fn = self.variant(update.periodic_update)

    for j in range(3):
      for i in range(1, update_period):
        new_params = update_fn(
            params_2, params_1, j*update_period+i, update_period)
        chex.assert_tree_all_close(
            params_1, new_params, atol=1e-10, rtol=1e-5)

      new_params = update_fn(
          params_2, params_1, (j+1)*update_period, update_period)
      chex.assert_tree_all_close(
          params_2, new_params, atol=1e-10, rtol=1e-5)


if __name__ == '__main__':
  absltest.main()



# Lint as: python3
# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
r"""Implementation of control variates.

 We are interested in computing the gradient using control variates:
  \nabla_{\theta} E_{p(x; \theta)} f(x)
    = \nabla_{\theta} [E_{p(x; \theta)} f(x) - h(x; \theta) + E_{p(x; \theta)}]
    = \nabla_{\theta} [E_{p(x; \theta)} f(x) - h(x; \theta)]
        + \nabla_{\theta} E_{p(x; \theta)}]
   = \nabla_{\theta} [E_{p(x; \theta)} f(x) - h(x; \theta)]
        + \nabla_{\theta} E_{p(x; \theta)}]
  = \nabla_{\theta} \int {p(x; \theta)} (f(x) - h(x; \theta)) dx
       + \nabla_{\theta} E_{p(x; \theta)}]
  =  \int \nabla_{\theta} {p(x; \theta)} (f(x) - h(x; \theta)) dx
       + [E_{p(x; \theta)} \nabla_{\theta} (f(x) - h(x; \theta))
       + \nabla_{\theta} E_{p(x; \theta)}]
  =  \int \nabla_{\theta} {p(x; \theta)} (f(x) - h(x; \theta)) dx
       - [E_{p(x; \theta)} \nabla_{\theta} h(x; \theta)
       + \nabla_{\theta} E_{p(x; \theta)}]

  The above computation is performed in `control_variates_jacobians`.

  When adding a new control variate, one does not need to implement the jacobian
  computation, but instead has to implement the forward computation.

  Each control variate implemented has to satisfy the following API:
    * control_variate(function)
        This returns a tuple of three functions:
           * The first element of the tuple is a function which returns the
                control variate value for a set of samples. It takes in as
                arguments the parameters used to construct the distribution,
                the distributional samples, and the state of the control variate
                (if any). The return value of this function will have shape
                 `num_samples`, where `num_samples` is the number of samples
                 provided as input.
           * The second is a function returns the expected value of the control
              variate. The input arguments of this function are the parameters
              of the distribution and the state of the control variate.
           * The third is a function which updates the state of the control
              variate, and returns the updated states.

  For examples, see `control_delta_method` and `moving_avg_baseline`.
"""
from typing import Any, Callable, Sequence, Tuple

import chex
import jax
import jax.numpy as jnp
from optax._src import transform


CvState = Any
ComputeCv = Callable[[transform.Params, chex.Array, CvState], float]
CvExpectedValue = Callable[[transform.Params, CvState], CvState]
UpdateCvState = Callable[[transform.Params, chex.Array, CvState], CvState]
ControlVariate = Tuple[ComputeCv, CvExpectedValue, UpdateCvState]


def control_delta_method(
    function: Callable[[chex.Array], float]) -> ControlVariate:
  """The control delta covariate method.

  Control variate obtained by performing a second order Taylor expansion
    on the cost function f at the mean of the input distribution.

  Only implemented for Gaussian random variables.

  For details, see: https://icml.cc/2012/papers/687.pdf

  Args:
    function: The function for which to compute the control variate.
      The function takes in one argument (a sample from the distribution) and
      returns a floating point value.

  Returns:
    A tuple of three functions, to compute the control variate, the
    expected value of the control variate, and to update the control variate
    state.
  """

  def delta(
      params: transform.Params,
      sample: chex.Array,
      state: CvState = None) -> chex.Array:
    """"Second order expansion of `function` at the mean of the input dist."""
    del state
    mean_dist = params[0]
    centered_sample = sample - mean_dist
    # Function is a function of samples. Here, we use the mean as the input
    # since we do a Taylor expansion of function around the mean.
    grads = jax.grad(function)(mean_dist)
    hessians = jax.hessian(function)(mean_dist)
    assert hessians.ndim == 2
    control_variate = function(mean_dist)
    control_variate += jnp.dot(centered_sample, grads)
    control_variate += jnp.dot(
        jnp.dot(centered_sample, hessians), centered_sample) / 2.
    return control_variate

  def expected_value_delta(
      params: transform.Params, state: CvState) -> float:
    """"Expected value of second order expansion of `function` at dist mean."""
    del state
    mean_dist = params[0]
    var_dist = jnp.square(jnp.exp(params[1]))
    hessians = jax.hessian(function)(mean_dist)

    assert hessians.ndim == 2
    hess_diags = jnp.diag(hessians)
    assert hess_diags.ndim == 1

    # Trace (Hessian * Sigma) and we use that Sigma is diagonal.
    expected_second_order_term = jnp.sum(var_dist * hess_diags) / 2.

    expected_control_variate = function(mean_dist)
    expected_control_variate += expected_second_order_term
    return expected_control_variate

  def update_state(
      params: transform.Params,
      samples: chex.Array,
      state: CvState = None) -> CvState:
    """"No state kept, so no operation is done."""
    del params, samples
    return state

  return delta, expected_value_delta, update_state


def moving_avg_baseline(
    function: Callable[[chex.Array], float],
    decay: float = 0.99,
    zero_debias: bool = True,
    use_decay_early_training_heuristic=True) -> ControlVariate:
  """A moving average baseline.

  It has no effect on the pathwise or measure valued estimator.

  Args:
    function: The function for which to compute the control variate.
      The function takes in one argument (a sample from the distribution) and
      returns a floating point value.
    decay: The decay rate for the moving average.
    zero_debias: Whether or not to use zero debiasing for the moving average.
    use_decay_early_training_heuristic: Whether or not to use a heuristic which
      overrides the decay value early in training based on
        min(decay, (1.0 + i) / (10.0 + i)). This stabilises training and was
        adapted from the Tensorflow codebase.

  Returns:
    A tuple of three functions, to compute the control variate, the
    expected value of the control variate, and to update the control variate
    state.
  """
  def moving_avg(
      params: transform.Params,
      samples: chex.Array,
      state: CvState = None) -> CvState:
    """"Return the moving average."""
    del params, samples
    return state[0]

  def expected_value_moving_avg(
      params: transform.Params, state: CvState) -> chex.Array:
    """"Return the moving average."""
    del params
    return state[0]

  def update_state(
      params: transform.Params,
      samples: chex.Array,
      state: CvState = None) -> CvState:
    """"Update the moving average."""
    del params
    value, i = state

    if use_decay_early_training_heuristic:
      iteration_decay = jnp.minimum(decay, (1.0 + i) / (10.0 + i))
    else:
      iteration_decay = decay

    updated_value = iteration_decay * value
    updated_value += (1 - iteration_decay) * jnp.mean(
        jax.vmap(function)(samples))

    if zero_debias:
      updated_value /= (jnp.ones([]) - jnp.power(iteration_decay, i + 1))

    return (jax.lax.stop_gradient(updated_value), i + 1)

  return moving_avg, expected_value_moving_avg, update_state


def _map(cv, params, samples, state):
  return jax.vmap(lambda x: cv(params, x, state))(samples)


def control_variates_jacobians(
    function: Callable[[chex.Array], float],
    control_variate_from_function: Callable[[Callable[[chex.Array], float]],
                                            ControlVariate],
    grad_estimator: Callable[..., jnp.array],
    params: transform.Params,
    dist_builder: Callable[..., Any],
    rng: chex.PRNGKey,
    num_samples: int,
    control_variate_state: CvState = None,
    estimate_cv_coeffs: bool = False,
    estimate_cv_coeffs_num_samples: int = 20) -> Tuple[
        Sequence[chex.Array], CvState]:
  r"""Obtain jacobians using control variates.

  We will compute each term individually. The first term will use stochastic
    gradient estimation. The second term will be computes using Monte
    Carlo estimation and automatic differentiation to compute
    \nabla_{\theta} h(x; \theta). The the third term will be computed using
    automatic differentiation, as we restrict ourselves to control variates
    which compute this expectation in closed form.

  This function updates the state of the control variate (once), before
    computing the control variate coefficients.

  Args:
    function: Function f(x) for which to estimate grads_{params} E_dist f(x).
      The function takes in one argument (a sample from the distribution) and
      returns a floating point value.
    control_variate_from_function: The control variate to use to reduce
      variance. See `control_delta_method` and `moving_avg_baseline` examples.
    grad_estimator: The gradient estimator to be used to compute the gradients.
      Note that not all control variates will reduce variance for all
      estimators. For example, the `moving_avg_baseline` will make no difference
      to the measure valued or pathwise estimators.
    params: A tuple of jnp arrays.
      The parameters for which to construct the distribution and for which we
      want to compute the jacobians.
    dist_builder: a constructor which builds a distribution given the input
      parameters specified by params. `dist_builder(params)` should return a
      valid distribution.
    rng: a PRNGKey key.
    num_samples: Int, the number of samples used to compute the grads.
    control_variate_state: The control variate state. This is used for control
      variates which keep states (such as the moving average baselines).
    estimate_cv_coeffs: Boolean. Whether or not to estimate the optimal control
      variate coefficient via `estimate_control_variate_coefficients`.
    estimate_cv_coeffs_num_samples: The number of samples to use to estimate
      the optimal coefficient. These need to be new samples to ensure that the
      objective is unbiased.

  Returns:
    A tuple of size two:
        * A tuple of size `params`, each element is `num_samples x param.shape`
            jacobian vector containing the estimates of the gradients obtained
            for each sample.
          The mean of this vector is the gradient wrt to parameters that can be
          used for learning. The entire jacobian vector can be used to assess
          estimator variance.
        * The updated CV state.
  """
  control_variate = control_variate_from_function(function)
  stochastic_cv, expected_value_cv, update_state_cv = control_variate
  data_dim = params[0].shape[0]
  if estimate_cv_coeffs:
    cv_coeffs = estimate_control_variate_coefficients(
        function, control_variate_from_function, grad_estimator, params,
        dist_builder, rng, estimate_cv_coeffs_num_samples,
        control_variate_state)
  else:
    cv_coeffs = [1.0] * len(params)

  # \int \nabla_{\theta} {p(x; \theta)} (f(x) - h(x; \theta)) dx
  function_jacobians = grad_estimator(
      function, params, dist_builder, rng, num_samples)

  # Chain rule since CVs can also depend on parameters - for example, for the
  # pathwise gradient estimator they have in order to have an effect on
  # gradient.
  # The rng has to be the same as passed to the grad_estimator above so that we
  # obtain the same samples.
  samples = dist_builder(*params).sample((num_samples,), seed=rng)
  # If the CV has state, update it.
  control_variate_state = update_state_cv(
      params, samples, control_variate_state)

  def samples_fn(x):
    return stochastic_cv(
        jax.lax.stop_gradient(params), x, control_variate_state)

  cv_jacobians = grad_estimator(
      samples_fn, params, dist_builder, rng, num_samples)

  # The gradients of the stochastic covariate with respect to the parameters.
  def param_fn(x):
    return jnp.mean(_map(
        stochastic_cv, x,
        jax.lax.stop_gradient(samples), control_variate_state))

  # [E_{p(x; \theta)} \nabla_{\theta} h(x; \theta)
  cv_param_grads = jax.grad(param_fn)(params)
  # The gradients of the closed form expectation of the control variate
  # with respect to the parameters: # \nabla_{\theta} E_{p(x; \theta)}].
  expected_value_grads = jax.grad(
      lambda x: expected_value_cv(x, control_variate_state))(params)

  jacobians = []
  for param_index, param in enumerate(params):
    chex.assert_shape(function_jacobians[param_index], (num_samples, data_dim))
    chex.assert_shape(cv_jacobians[param_index], (num_samples, data_dim))
    chex.assert_shape(cv_param_grads[param_index], (data_dim,))
    chex.assert_shape(expected_value_grads[param_index], (data_dim,))

    cv_coeff = cv_coeffs[param_index]
    # \int \nabla_{\theta} {p(x; \theta)} (f(x) - h(x; \theta)) dx
    param_jacobians = function_jacobians[param_index]
    param_jacobians -= cv_coeff * cv_jacobians[param_index]
    # - [E_{p(x; \theta)} \nabla_{\theta} h(x; \theta)
    param_jacobians -= cv_coeff * cv_param_grads[param_index]
    # \nabla_{\theta} E_{p(x; \theta)}]
    param_jacobians += cv_coeff * expected_value_grads[param_index]

    chex.assert_shape(param_jacobians, (num_samples,) + param.shape)
    jacobians.append(param_jacobians)

  return jacobians, control_variate_state


def estimate_control_variate_coefficients(
    function: Callable[[chex.Array], float],
    control_variate_from_function: Callable[[Callable[[chex.Array], float]],
                                            ControlVariate],
    grad_estimator: Callable[..., jnp.array],
    params: transform.Params,
    dist_builder: Callable[..., Any],
    rng: chex.PRNGKey,
    num_samples: int,
    control_variate_state: CvState = None,
    eps: float = 1e-3) -> Sequence[float]:
  r"""Estimates the control variate coefficients for the given parameters.

  For each variable `var_k`, the coefficient is given by:
    \sum_k cov(df/d var_k, d cv/d var_k) / (\sum var(d cv/d var_k) + eps)

  Where var_k is the k'th element of the parameters in `params`.
  The covariance and variance calculations are done from samples obtained
  from the distribution obtained by calling `dist_builder` on the input
  `params`.

  This function does not update the state of the control variate.

  Args:
    function: Function f(x) for which to estimate grads_{params} E_dist f(x).
      The function takes in one argument (a sample from the distribution) and
      returns a floating point value.
    control_variate_from_function: The control variate to use to reduce
      variance. See `control_delta_method` and `moving_avg_baseline` examples.
    grad_estimator: The gradient estimator to be used to compute the gradients.
      Note that not all control variates will reduce variance for all
      estimators. For example, the `moving_avg_baseline` will make no difference
      to the measure valued or pathwise estimators.
    params: A tuple of jnp arrays.
      The parameters for which to construct the distribution and for which we
      want to compute the jacobians.
    dist_builder: a constructor which builds a distribution given the input
      parameters specified by params. `dist_builder(params)` should return a
      valid distribution.
    rng: a PRNGKey key.
    num_samples: Int, the number of samples used to compute the grads.
    control_variate_state: The control variate state. This is used for control
      variates which keep states (such as the moving average baselines).
    eps: A small constant used to avoid numerical issues. Float.

  Returns:
    A list of control variate coefficients (each a scalar), for each parameter
    in `params`.
  """
  # Resample to avoid biased gradients.
  cv_rng, _ = jax.random.split(rng)
  del rng  # Avoid using rng in this function.
  stochastic_cv, _, _ = control_variate_from_function(function)

  # Samples have to be the same so we use the same rng.
  cv_jacobians = grad_estimator(
      lambda x: stochastic_cv(params, x, control_variate_state),
      params, dist_builder, cv_rng, num_samples)
  function_jacobians = grad_estimator(
      function, params, dist_builder, cv_rng, num_samples)

  def compute_coeff(param_cv_jacs, param_f_jacs):
    assert param_f_jacs.ndim == 2
    assert param_cv_jacs.ndim == 2

    mean_f = jnp.mean(param_f_jacs, axis=0)
    mean_cv = jnp.mean(param_cv_jacs, axis=0)

    cov = jnp.mean((param_f_jacs - mean_f) * (param_cv_jacs - mean_cv), axis=0)

    assert cov.ndim == 1

    # Compute the coefficients which minimize variance.
    # Since we want to minimize the variances across parameter dimensions,
    # the optimal coefficients are given by the sum of covariances per
    # dimensions over the sum of variances per dimension.
    cv_coeff = jnp.sum(cov) / (jnp.sum(jnp.var(param_cv_jacs, axis=0)) + eps)
    return jax.lax.stop_gradient(cv_coeff)

  return [compute_coeff(cv_jacobians[i], function_jacobians[i])
          for i in range(len(params))]
